#include "typedef.h"

#include <assert.h>
#include <string.h>
#include <stdio.h>

#include "stringcache.h"
#include "object.h"
#include "xmalloc.h"
#include "opcode.h"

#include "code.h"
#include "hash.h"
#include "file.h"
#include "source.h"
#include "class.h"
#include "function.h"
#include "sort.h"
#include "dex.h"

extern FILE	*stdoutF;

extern int	g_verbose;
extern u4T	g_max_function_lth;

extern	u1T	*g_imageP;
extern	int	g_image_size;

// static for efficiency
static Csource		*g_sourceP;
static Cclass		*g_classP;
static Cfunction	*g_functionP;

static u4T			g_strings;
static stringT		*g_stringsP;
static u4T			g_types;
static stringT		**g_typesPP;
static u4T			g_protos;
static protoT		*g_protosP;
static u4T			g_fields;
static fieldT		*g_fieldsP;
static u4T			g_methods;
static methodT		*g_methodsP;

static Cheader_item	*g_headerP;
static u4T			g_data_end;


/* static */ void
Cdexerror::error(const char *messageP)
{
	throw new Cdexerror(messageP);
}

static void
in_data_section(u4T offset)
{
	if (offset < g_headerP->data_off) {
		Cdexerror::error("Offset before data section");
	}
	if (g_data_end <= offset) {
		Cdexerror::error("Offset after data section");
	}
}

static stringT	*
getString(u4T index)
{
	if (index == NO_INDEX) {
		return(0);
	}
	if (g_strings <= index) {
		Cdexerror::error("Illegal string index");
	}
	return(g_stringsP + index);
}

static stringT	*
getType(u4T index)
{
	if (index == NO_INDEX) {
		return(0);
	}
	if (g_types <= index) {
		Cdexerror::error("Illegal type index");
	}
	return(g_typesPP[index]);
}

static protoT	*
getProto(u4T index)
{
	if (index == NO_INDEX) {
		return(0);
	}
	if (g_protos <= index) {
		Cdexerror::error("Illegal proto index");
	}
	return(g_protosP + index);
}

static fieldT	*
getField(u4T index)
{
	if (index == NO_INDEX) {
		return(0);
	}
	if (g_fields <= index) {
		Cdexerror::error("Illegal field index");
	}
	return(g_fieldsP + index);
}

static methodT	*
getMethod(u4T index)
{
	extern int	g_dynamic;
	methodT		*retP;

	if (index == NO_INDEX) {
		return(0);
	}
	if (g_methods <= index) {
		Cdexerror::error("Illegal method index");
	}
	retP = g_methodsP + index;
	if (g_dynamic) {
		Cclass::locate(retP->classP);
	}
	return(retP);
}

static void
fixup_uint(u1T *fixP)
{
	register u1T	u1;

	u1      = fixP[0];
	fixP[0] = fixP[3];
	fixP[3] = u1;
	u1      = fixP[1];
	fixP[1] = fixP[2];
	fixP[2] = u1;
}

static void
fixup_ushort(u1T *fixP)
{
	register	u1T u1;

	u1      = fixP[0];
	fixP[0] = fixP[1];
	fixP[1] = u1;
}

/*
	http://www.netmite.com/android/mydroid/dalvik/docs/dex-format.html

	LEB128
	LEB128 ("Little-Endian Base 128") is a variable-length encoding for arbitrary signed or unsigned integer quantities. 
	The format was borrowed from the DWARF3 specification. 
	In a .dex file, LEB128 is only ever used to encode 32-bit quantities.

	Each LEB128 encoded value consists of one to five bytes, which together represent a single 32-bit value. 
	Each byte has its most significant bit set except for the final byte in the sequence, which has its most significant bit clear. 
	The remaining seven bits of each byte are payload, with the least significant seven bits of the quantity in the first byte, 
	the next seven in the second byte and so on. 
	In the case of a signed LEB128 (sleb128), the most significant payload bit of the final byte in the sequence is sign-extended to produce the final value. 
	In the unsigned case (uleb128), any bits not explicitly represented are interpreted as 0. 

	Bitwise diagram of a two-byte LEB128 value 
	First byte                           Second byte  
	1 bit6 bit5 bit4 bit3 bit2 bit1 bit0 0 bit13 bit12 bit11 bit10 bit9 bit8 bit7 

	The variant uleb128p1 is used to represent a signed value, where the representation is of the value plus one encoded as a uleb128. 
	This makes the encoding of -1 (alternatively thought of as the unsigned value 0xffffffff) 
	 but no other negative number  a single byte, and is useful in exactly those cases where the represented number must either be non-negative 
	or -1 (or 0xffffffff), and where no other negative values are allowed (or where large unsigned values are unlikely to be needed).

	Here are some examples of the formats:

	Encoded Sequence As sleb128 As uleb128 As uleb128p1 
	00                  0          0          -1 
    01                  1          1           0 
    7f                  -1         127         126 
    80 7f               -128       16256       16255 
*/

static u4T
decode_uleb128(u1T **bufferPP)
{
	u1T	*bufferP;
	u8T	ret, temp;
	s4T	shift;
	u4T	c;


	ret     = 0;
	shift   = 0;
	bufferP = *bufferPP;

	for (;;) {
		c = *bufferP++;
		assert(shift < 64);
		temp   = (c & 0x7F);
		temp <<= shift;
		ret   |= temp;
		if (!(c & 0x80)) {
			break;
		}
		shift += 7;
	}
	*bufferPP = bufferP;
	assert(ret <= 0xFFFFFFFF);
	return((u4T) ret);
}

static s4T
decode_sleb128(u1T **bufferPP)
{
	u1T	*bufferP;
	s8T	ret;
	u8T	temp;
	s4T	shift;
	u4T	c;


	ret     = 0;
	shift   = 0;
	bufferP = *bufferPP;

	do {
		c = *bufferP++;
		assert(shift < 64);
		temp   = (c & 0x7F);
		temp <<= shift;
		ret   |= temp;
		shift += 7;
	} while (c & 0x80);

	if ((shift < 64) && (c & 0x40)) {
		temp   = 1;
		temp <<= shift;
		// Sign extend
		ret   |= (u8T) -((s8T) temp);
	}
	*bufferPP = bufferP;
	assert(ret <= 0x7FFFFFFF);
	return((s4T) ret);
}

/* String Syntax

	There are several kinds of item in a .dex file which ultimately refer to a string. 
	The following BNF-style definitions indicate the acceptable syntax for these strings.

	SimpleName
	A SimpleName is the basis for the syntax of the names of other things. 
	The .dex format allows a fair amount of latitude here (much more than most common source languages). 
	In brief, a simple name may consist of any low-ASCII alphabetic character or digit, a few specific low-ASCII symbols, and most non-ASCII code points that are not control, space, or special characters. 
	Note that surrogate code points (in the range U+d800  U+dfff) are not considered valid name characters, per se, 
	but Unicode supplemental characters are valid (which are represented by the final alternative of the rule for SimpleNameChar), 
	and they should be represented in a file as pairs of surrogate code points in the MUTF-8 encoding.

	SimpleName ? SimpleNameChar (SimpleNameChar)* 
	SimpleNameChar ? 
		  'A'  'Z' 
		| 'a'  'z' 
		| '0'  '9' 
		| '$' 
		| '-' 
		| '_' 
		| U+00a1  U+1fff 
		| U+2010  U+2027 
		| U+2030  U+d7ff 
		| U+e000  U+ffef 
		| U+10000  U+10ffff 

	MemberName used by field_id_item and method_id_item
	A MemberName is the name of a member of a class, members being fields, methods, and inner classes.

	MemberName ?  SimpleName | '<' SimpleName '>' 

	FullClassName
	A FullClassName is a fully-qualified class name, including an optional package specifier followed by a required name.

	FullClassName ?  OptionalPackagePrefix SimpleName 
	OptionalPackagePrefix ?  (SimpleName '/')* 
*/

// string identifiers list. 
// These are identifiers for all the strings used by this file, either for internal naming (e.g., type descriptors) or as constant objects referred to by code. 
// This list must be sorted by string contents, using UTF-16 code point values (not in a locale-sensitive manner). 

/* MUTF-8 (Modified UTF-8) Encoding

	As a concession to easier legacy support, the .dex format encodes its string data in a de facto standard modified UTF-8 form, hereafter referred to as MUTF-8. 
	This form is identical to standard UTF-8, except:

	 Only the one-, two-, and three-byte encodings are used.
	 Code points in the range U+10000  U+10ffff are encoded as a surrogate pair, each of which is represented as a three-byte encoded value.
	 The code point U+0000 is encoded in two-byte form.
	 A plain null byte (value 0) indicates the end of a string, as is the standard C language interpretation.

	The first two items above can be summarized as: MUTF-8 is an encoding format for UTF-16, instead of being a more direct encoding format for Unicode characters.

	The final two items above make it simultaneously possible to include the code point U+0000 in a string and still manipulate it as a C-style null-terminated string.

	However, the special encoding of U+0000 means that, unlike normal UTF-8, the result of calling the standard C function strcmp() on a pair of MUTF-8 strings does not always indicate the properly signed result of comparison of unequal strings. 
	When ordering (not just equality) is a concern, the most straightforward way to compare MUTF-8 strings is to decode them character by character, and compare the decoded values. 
	(However, more clever implementations are also possible.)

	Please refer to The Unicode Standard for further information about character encoding. MUTF-8 is actually closer to the (relatively less well-known) encoding CESU-8 than to UTF-8 per se.
*/
/* static */ void
Cstring_id_item::check_offset(u4T size, u4T offset)
{
	if (!size) {
		if (offset) {
			Cdexerror::error("Invalid string_ids_off -- should be 0");
		}
		return;
	}
	if (offset < g_headerP->header_size) {
		Cdexerror::error("Invalid string_ids_off -- in header");
	}
	if (offset & 3) {
		Cdexerror::error("Invalid string_ids_off -- not integer aligned");
}	}

/* static */ void
Cstring_id_item::fixup_all(u4T size, u4T offset)
{
	u4T		*string_data_offP, *endP;

	check_offset(size, offset);
	string_data_offP = (u4T *) (g_imageP + offset);
	for (endP = string_data_offP + size; string_data_offP < endP; ++string_data_offP) {
		// uint offset from the start of the file to the string data for this item. 
		// The offset should be to a location in the data section, and the data should be in the format specified by "string_data_item" below. 
		// There is no alignment requirement for the offset.  
		fixup_uint((u1T *) string_data_offP);
}	}

/*
  4.4.7 The CONSTANT_Utf8_info Structure

  The CONSTANT_Utf8_info structure is used to represent constant string values. UTF-8 strings are encoded so that character 
  sequences that contain only non-null ASCII characters can be represented using only 1 byte per character, but characters 
  of up to 16 bits can be represented. All characters in the range '\u0001' to '\u007F' are represented by a single byte: 

0  bits 6-0  

The 7 bits of data in the byte give the value of the character represented. 
The null character ('\u0000') and characters in the range '\u0080' to '\u07FF' are represented by a pair of bytes x and y:

x:  1  1  0  bits 10-6  
y:  1  0  bits 5-0  

The bytes represent the character with the value ((x & 0x1f) << 6) + (y & 0x3f). 

Characters in the range '\u0800' to '\uFFFF' are represented by 3 bytes x, y, and z:

x:  1  1  1  0  bits 15-12  
y:  1  0  bits 11-6  
z:  1  0  bits 5-0  

The character with the value ((x & 0xf) << 12) + ((y & 0x3f) << 6) + (z & 0x3f) is represented by the bytes. 

The bytes of multibyte characters are stored in the class file in big-endian (high byte first) order.

*/

static void
getUtf8(const u1T *compressedP, u2T **stringPP, int load)
{
	const u1T	*P       = compressedP;
	u2T			*startP  = *stringPP;
	u2T			*stringP = startP;
	u1T			c;
	u2T			val;

	for (; (c = *P); ++P) {
		if (!(c & 0x80)) {
			// One byte encoding
			val = c;
		} else {
			if (!(c & 0x20)) {
				// Two byte encoding
				val  = (c & 0x1F) << 6;
			} else {
				// Three byte encoding
				val  = (c & 0xF) << 12;
				c    = *++P;
				val |= (c & 0x3F) << 6;
			}
			c    = *++P;
			val |= (c & 0x3F);
		}
		if (load) {
			*stringP = val;
		}
		++stringP;
	}
	if (load) {
		*stringP = 0;
	}
	*stringPP  = ++stringP;
}

/* static */ void
Cstring_id_item::scan_all(u4T size, u4T offset)
{
	u4T		lth;
	u4T		*string_data_offP, *endP;
	u4T		string_data_off;
	stringT	*stringP;
	u2T		*startP, *targetP;
	u1T		*P;
	int		load;

	check_offset(size, offset);
	if (!size) {
		return;
	}
	g_sourceP->m_strings  = g_strings  = size;
	g_sourceP->m_stringsP = g_stringsP = stringP = (stringT *) Xmalloc(size * sizeof(stringT));

	targetP = 0;
	for (load = 0; ; ++load) {
		string_data_offP = (u4T *) (g_imageP + offset);
		for (endP = string_data_offP + size; string_data_offP < endP; ) {
			// uint offset from the start of the file to the string data for this item. 
			// The offset should be to a location in the data section, and the data should be in the format specified by "string_data_item" below. 
			// There is no alignment requirement for the offset.  
			string_data_off = *string_data_offP++; 
			in_data_section(string_data_off);
			P = g_imageP + string_data_off;
			lth = decode_uleb128(&P);			// This is the UFT16 string length
			startP = targetP;
			getUtf8(P, &targetP, load);
			if (load) {
				stringP->lth = (targetP - startP) - 1;
				assert(stringP->lth == lth);
				stringP->P   = stringCache(startP);
				++stringP;
		}	}
		if (load) {
			break;
		}
		targetP = (u2T *) Xmalloc(((u1T *) targetP) - (u1T *) 0);
	}
}

// type identifiers list. 
// These are identifiers for all types (classes, arrays, or primitive types) referred to by this file, whether defined in the file or not. 
// This list must be sorted by string_id index.

/*	TypeDescriptor used by type_id_item
	A TypeDescriptor is the representation of any type, including primitives, classes, arrays, and void. 
	See below for the meaning of the various versions.

	TypeDescriptor ? 
		 'V'					// void; only valid for return types
		| FieldTypeDescriptor 
	FieldTypeDescriptor ? 
		  NonArrayFieldTypeDescriptor 
		| ('[' * 1255) NonArrayFieldTypeDescriptor 
								// array of descriptor, usable recursively for arrays-of-arrays, though it is invalid to have more than 255 dimensions
	NonArrayFieldTypeDescriptor? 
		  'Z'					// boolean
		| 'B'					// byte
		| 'S'					// short
		| 'C'					// char
		| 'I'					// int
		| 'J'					// long
		| 'F'					// float
		| 'D'					// double
		| 'L' FullClassName ';'	// the class fully.qualified.Name
*/

/* static */ void
Ctype_id_item::check_offset(u4T size, u4T offset)
{
	if (!size) {
		if (offset) {
			Cdexerror::error("Invalid type_ids_off -- should be 0");
		}
		return;
	}
	if (offset < g_headerP->header_size) {
		Cdexerror::error("Invalid type_ids_off -- in header");
	}
	if (offset & 3) {
		Cdexerror::error("Invalid type_ids_off -- not integer aligned");
}	}

/* static */ void
Ctype_id_item::fixup_all(u4T size, u4T offset)
{
	u4T		*typeP, *endP;

	check_offset(size, offset);

	typeP = (u4T *) (g_imageP + offset);
	for (endP = typeP + size; typeP < endP; ++typeP) {
		// index into the string_ids list for the descriptor string of this type. 
		// The string must conform to the syntax for TypeDescriptor, defined above.
		fixup_uint((u1T *) typeP);
	}
	return;
}

/* static */ void
Ctype_id_item::scan_all(u4T size, u4T offset)
{
	u4T		type, *typeP, *end_typesP;
	stringT	**targetPP;

	if (!size) {
		if (offset) {
			Cdexerror::error("Invalid type_ids_off -- should be 0");
		}
		return;
	}
	if (offset < g_headerP->header_size) {
		Cdexerror::error("Invalid type_ids_off -- in header");
	}
	if (offset & 3) {
		Cdexerror::error("Invalid type_ids_off -- not integer aligned");
	}
	typeP                = (u4T *) (g_imageP + offset);
	g_sourceP->m_types   = g_types = size;
	g_sourceP->m_typesPP = g_typesPP = targetPP = (stringT **) Xmalloc(size * sizeof(stringT *));
	for (end_typesP = typeP + size; typeP < end_typesP; ) {
		type = *typeP++;
		// index into the string_ids list for the descriptor string of this type. 
		// The string must conform to the syntax for TypeDescriptor, defined above.
		*targetPP++ = getString(type);
	}
	return;
}

/* ShortyDescriptor used by proto_id_item

	A ShortyDescriptor is the short form representation of a method prototype, including return and parameter types, 
	except that there is no distinction between various reference (class or array) types. 
	Instead, all reference types are represented by a single 'L' character.

	ShortyDescriptor ? 
		ShortyReturnType (ShortyFieldType)* 
	ShortyReturnType ? 
		  'V'				// void; only valid for return types
		| ShortyFieldType 
	ShortyFieldType ? 
		  'Z'				// boolean
		| 'B'				// byte
		| 'S'				// short
		| 'C'				// char
		| 'I'				// int
		| 'J'				// long 
		| 'F'				// float
		| 'D'				// double
		| 'L'				// reference 
*/

/* static */ void
Ctype_list::check_offset(u4T offset)
{
	if (offset & 3) {
		Cdexerror::error("Invalid type_list -- not aligned");
	}
	in_data_section(offset);
}

/* static */ void
Ctype_list::fixup(u4T offset)
{
	u4T		*type_listP;
	u4T		i;
	u2T		*type_itemP, *endP;

	if (offset) {
		check_offset(offset);
		type_listP = (u4T *) (g_imageP + offset);
		// size uint size of the list, in entries 
		// list type_item[size] elements of the list 

		fixup_uint((u1T *) type_listP);
		if ((i = *type_listP)) {
			type_itemP = (u2T *) (type_listP + 1);
			for (endP = type_itemP + i; type_itemP < endP; ++type_itemP) {
				fixup_ushort((u1T *) type_itemP);
	}	}	}
	return;
}

/* static */ int
Ctype_list::scan(u4T offset, stringT ***PPP, int load)
{
	u4T		*type_listP;
	stringT	**PP;
	u4T		size;
	u2T		type, *type_itemP, *endP;

	if (!offset) {
		return(0);
	}
	check_offset(offset);
	type_listP = (u4T *) (g_imageP + offset);
	PP         = *PPP;

	// size uint size of the list, in entries 
	// list type_item[size] elements of the list 

	if ((size = *type_listP)) {
		type_itemP = (u2T *) (type_listP + 1);

		for (endP = type_itemP + size; type_itemP < endP; ) {
			type = *type_itemP++;
			if (load) {
				*PP = getType(type); 
			}
			++PP;
	}	}
	*PPP = PP;
	return(size);
}

/* static */ void
Cproto_id_item::check_offset(u4T size, u4T offset)
{
	if (!size) {
		if (offset) {
			Cdexerror::error("Invalid proto_ids_off -- should be 0");
		}
		return;
	}
	if (offset < g_headerP->header_size) {
		Cdexerror::error("Invalid proto_ids_off -- in header");
	}
	if (offset & 3) {
		Cdexerror::error("Invalid proto_ids_off -- not integer aligned");
}	}

/* static */ void
Cproto_id_item::fixup_all(u4T size, u4T offset)
{
	Cproto_id_item	*protoP, *endP;

	check_offset(size, offset);
	if (!size) {
		return;
	}
	protoP = (Cproto_id_item *) (g_imageP + offset);
	for (endP = protoP + size; protoP < endP; ++protoP) {
		fixup_uint((u1T *) &protoP->shorty_idx);
		fixup_uint((u1T *) &protoP->return_type_idx);
		fixup_uint((u1T *) &protoP->parameters_off);
		Ctype_list::fixup(protoP->parameters_off);
}	}

/* static */ void
Cproto_id_item::scan_all(u4T size, u4T offset)
{
	u4T				need;
	u4T				parameters;
	Cproto_id_item	*protosP, *end_protosP, *protoP;
	u4T				*type_listP;
	u2T				*parameter_listP;
	u4T				parameters_offset;
	protoT			*targetP;
	stringT			**PP;
	stringT			**parametersPP, **end_parametersPP;

	check_offset(size, offset);
	if (!size) {
		return;
	}

	PP         = 0;
	parameters = 0;
	protoP     = protosP = (Cproto_id_item *) (g_imageP + offset);
	for (end_protosP = protosP + size; protoP < end_protosP; ++protoP) {
		parameters_offset = protoP->parameters_off;
		if (parameters_offset) {
			Ctype_list::check_offset(parameters_offset);
			type_listP  = (u4T *) (g_imageP + parameters_offset);
			parameters += *type_listP;
	}	}

	need                 = (size * sizeof(protoT)) + (parameters * sizeof(stringT *));
	g_sourceP->m_protos  = g_protos  = size;
	g_sourceP->m_protosP = g_protosP = targetP = (protoT *) Xmalloc(need);
	parametersPP         = (stringT **) (targetP + size);

	for (protoP = protosP; protoP < end_protosP; ++protoP) {
		targetP->shortyP      = getString(protoP->shorty_idx);
		targetP->return_typeP = getType(protoP->return_type_idx);
		parameters_offset     = protoP->parameters_off;
		if (!parameters_offset) {
			targetP->parameters   = 0;
			targetP->parametersPP = 0;
		} else {
			type_listP            = (u4T *) (g_imageP + parameters_offset);
			parameter_listP       = (u2T *) (type_listP + 1);
			targetP->parameters   = parameters = *type_listP;
			targetP->parametersPP = parametersPP;
			for (end_parametersPP = parametersPP + parameters; parametersPP < end_parametersPP; ) {
				offset            = *parameter_listP++;
				*parametersPP++   = getType(offset);
		}	}	
		++targetP;
	}
	assert(parametersPP == (stringT **) (((u1T *) g_sourceP->m_protosP) + need));
}

/* static */ void
Cfield_id_item::check_offset(u4T size, u4T offset)
{
	if (!size) {
		if (offset) {
			Cdexerror::error("Invalid field_ids_off -- should be 0");
		}
		return;
	}
	if (offset < g_headerP->header_size) {
		Cdexerror::error("Invalid field_ids_off -- in header");
	}
	if (offset & 3) {
		Cdexerror::error("Invalid field_ids_off -- not integer aligned");
}	}

/* static */ void
Cfield_id_item::fixup_all(u4T size, u4T offset)
{
	Cfield_id_item	*fieldP, *endP;

	check_offset(size, offset);
	fieldP = (Cfield_id_item *) (g_imageP + offset);
	for (endP = fieldP + size; fieldP < endP; ++fieldP) {
		fixup_ushort((u1T *) &fieldP->class_idx);
		fixup_ushort((u1T *) &fieldP->type_idx);
		fixup_uint((u1T *)   &fieldP->name_idx);
}	}

/* static */ void
Cfield_id_item::scan_all(u4T size, u4T offset)
{
	u4T				need;
	Cfield_id_item	*fieldP, *end_fieldsP;
	fieldT			*targetP;

	check_offset(size, offset);
	fieldP               = (Cfield_id_item *) (g_imageP + offset);
	need                 = size * sizeof(fieldT);
	g_sourceP->m_fields  = g_fields  = size;
	g_sourceP->m_fieldsP = targetP = g_fieldsP = (fieldT *) Xmalloc(need);
	for (end_fieldsP = fieldP + size; fieldP < end_fieldsP; ++fieldP) {
		targetP->classP      = getType(fieldP->class_idx);
		targetP->typeP       = getType(fieldP->type_idx);
		targetP->nameP       = getString(fieldP->name_idx);
		++targetP;
}	}

/* static */ void
Cmethod_id_item::check_offset(u4T size, u4T offset)
{
	if (!size) {
		if (offset) {
			Cdexerror::error("Invalid method_ids_off -- should be 0");
		}
		return;
	}
	if (offset < g_headerP->header_size) {
		Cdexerror::error("Invalid method_ids_off -- in header");
	}
	if (offset & 3) {
		Cdexerror::error("Invalid method_ids_off -- not integer aligned");
	}
}

/* static */ void
Cmethod_id_item::fixup_all(u4T size, u4T offset)
{
	Cmethod_id_item	*methodP, *endP;

	check_offset(size, offset);

	methodP = (Cmethod_id_item *) (g_imageP + offset);
	for (endP = methodP + size; methodP < endP; ++methodP) {
		fixup_ushort((u1T *) &methodP->class_idx);
		fixup_ushort((u1T *) &methodP->proto_idx);
		fixup_uint((u1T *)   &methodP->name_idx);
}	}

/* static */ void
Cmethod_id_item::scan_all(u4T size, u4T offset)
{
	Cmethod_id_item	*methodP, *endP;
	methodT			*targetP;

	check_offset(size, offset);

	methodP               = (Cmethod_id_item *) (g_imageP + offset);
	g_sourceP->m_methods  = g_methods = size;
	g_sourceP->m_methodsP = targetP = g_methodsP = (methodT *) Xmalloc(size * sizeof(methodT));
	for (endP = methodP + size; methodP < endP; ++methodP) {
		targetP->classP      = getType(methodP->class_idx);
		targetP->protoP      = getProto(methodP->proto_idx);
		targetP->nameP       = getString(methodP->name_idx);
		++targetP;
}	}

// ==========================================================

static s4T
decode_s4(int lth, u1T *P)
{
	u1T	*endP;
	s4T	val;
	int shift;

	val = 0;
	for (endP = P + lth; P < endP; ) {
		val <<= 8;
		val  |= *(--endP);
	}
	if ((shift = (4 - lth))) {
		shift <<= 3;
		val   <<= shift;
		val   >>= shift;
	}
	return(val);
}

static u4T
decode_u4(int lth, u1T *P)
{
	u1T	*endP;
	u4T	val;

	val = 0;
	for (endP = P + lth; P < endP; ) {
		val <<= 8;
		val  |= *(--endP);
	}
	return(val);
}

static s8T
decode_s8(int lth, u1T *P)
{
	u1T	*endP;
	s8T	val;
	int	shift;

	val = 0;
	for (endP = P + lth; P < endP; ) {
		val <<= 8;
		val  |= *(--endP);
	}
	if ((shift = (8 - lth))) { 
		shift <<= 3;
		val   <<= shift;
		val   >>= shift;
	}
	return(val);
}

static u4T
decode_float(int lth, u1T *P)
{
	u1T	*endP;
	u4T	val;
	int	shift;
	
	val = 0;
	for (endP = P + lth; P < endP; ) {
		val <<= 8;
		val  |= *(--endP);
	}
	if ((shift = (4 - lth))) {
		val <<= (shift << 3);
	}
	return(val);
}

static u8T
decode_double(int lth, u1T *P)
{
	u1T	*endP;
	u8T	val;
	int	shift;
	
	val = 0;
	for (endP = P + lth; P < endP; ) {
		val <<= 8;
		val  |= *(--endP);
	}
	if ((shift = (8 - lth))) {
		val <<= (shift << 3);
	}
	return(val);
}

static void 
encoded_array(u1T **PP, annotation_arrayT **targetPP, u1T **memoryPP, int load)
{
	u4T					size;
	annotation_arrayT	*arrayP;
	annotation_valueT	*entryP, *end_entryP;

	size   = decode_uleb128(PP);
	arrayP = (annotation_arrayT *) *memoryPP;

	if (load) {
		arrayP->size = size;
		*targetPP    = arrayP;
	}
	entryP             = arrayP->value;
	end_entryP         = entryP + size;
	*memoryPP          = (u1T *) end_entryP;

	for (; entryP < end_entryP; ++entryP) {
		Cencoded_value::scan(PP, entryP, memoryPP, load);
	}
	return;
}

/* static */ void
Cencoded_value::scan(u1T **PP, annotation_valueT *targetP, u1T **memoryPP, int load)
{
	u1T	*P;
	u1T	value;
	u1T	value_type;
	u1T	value_arg;

	P          = *PP;
	value      = *P++;
	value_type = value & 0x1F;
	value_arg  = value >> 5;

	if (load) {
		targetP->type = (annotation_valueE) value_type;
	}

	switch (value_type) {
	case VALUE_BYTE:		// 0x00	(none; must be 0) ubyte[1] signed one-byte integer value
		if (value_arg) {
			Cdexerror::error("Invalid VALUE_BYTE");
		}
		if (load) {
			targetP->value._s1 = * ((s1T *) P);
		}
		++P;
		break;
	case VALUE_SHORT:		// 0x02	size - 1 (01) ubyte[size] signed two-byte integer value, sign-extended
		switch (value_arg) {
		case 0:
			if (load) {
				targetP->value._s2 = *((s1T *) P);
			}
			++P;
			break;
		case 1:
			if (load) {
				memcpy(&targetP->value._s2, P, sizeof(s2T));
			}
			P += 2;
			break;
		default:
			Cdexerror::error("Invalid VALUE_SHORT");
		}
		break;
	case VALUE_CHAR:		// 0x03	size - 1 (01) ubyte[size] unsigned two-byte integer value, zero-extended 
		switch (value_arg) {
		case 0:
			if (load) {
				targetP->value._u2 = *((u1T *) P);
			}
			++P;
			break;
		case 1:
			if (load) {
				memcpy(&targetP->value._u2, P, sizeof(u2T));
			}
			P += 2;
			break;
		default:
			Cdexerror::error("Invalid VALUE_CHAR");
		}
		break;
	case VALUE_INT:			// 0x04	size - 1 (03) ubyte[size] signed four-byte integer value, sign-extended
		if (value_arg > 3) {
			Cdexerror::error("Invalid VALUE_INT");
		}
		++value_arg;
		if (load) {
			targetP->value._s4 = decode_s4(value_arg, P);
		}
		P += value_arg;
		break;
	case VALUE_LONG:		// 0x06	size - 1 (07) ubyte[size] signed eight-byte integer value, sign-extended
		if (value_arg > 7) {
			Cdexerror::error("Invalid VALUE_LONG");
		}
		++value_arg;
		if (load) {
			targetP->value._s8 = decode_s8(value_arg, P);
		}
		P += value_arg;
		break;
	case VALUE_FLOAT:		// 0x10	size - 1 (03) ubyte[size] four-byte bit pattern, zero-extended to the right, and interpreted as an IEEE754 32-bit floating point value  
		if (value_arg > 3) {
			Cdexerror::error("Invalid VALUE_FLOAT");
		}
		++value_arg;
		if (load) {
			targetP->value._u4 = decode_float(value_arg, P);
		}
		P += value_arg;
		break;
	case VALUE_DOUBLE:		// 0x11	size - 1 (07) ubyte[size] eight-byte bit pattern, zero-extended to the right, and interpreted as an IEEE754 64-bit floating point value  
		if (value_arg > 7) {
			Cdexerror::error("Invalid VALUE_DOUBLE");
		}
		++value_arg;
		if (load) {
			targetP->value._u8 = decode_double(value_arg, P);
		}
		P += value_arg;
		break;
	case VALUE_STRING:		// 0x17	size - 1 (03) ubyte[size] unsigned (zero-extended) four-byte integer value, interpreted as an index into the string_ids section and representing a string value  
		if (value_arg > 3) {
			Cdexerror::error("Invalid VALUE_STRING");
		}
		++value_arg;
		if (load) {
			u4T	val;
			
			val = decode_u4(value_arg, P);
			targetP->value._stringP = getString(val);
		}
		P += value_arg;
		break;
	case VALUE_TYPE:		// 0x18	size - 1 (03) ubyte[size] unsigned (zero-extended) four-byte integer value, interpreted as an index into the type_ids section and representing a reflective type/class value  
		if (value_arg > 3) {
			Cdexerror::error("Invalid VALUE_TYPE");
		}
		++value_arg;
		if (load) {
			u4T	val;
	
			val = decode_u4(value_arg, P);
			targetP->value._stringP = getType(val);
		}
		P += value_arg;
		break;
	case VALUE_FIELD:		// 0x19	size - 1 (03) ubyte[size] unsigned (zero-extended) four-byte integer value, interpreted as an index into the field_ids section and representing a reflective field value  
		if (value_arg > 3) {
			Cdexerror::error("Invalid VALUE_FIELD");
		}
		++value_arg;
		if (load) {
			u4T	val;
			
			val = decode_u4(value_arg, P);
			if (g_sourceP->m_fields <= val) {
				Cdexerror::error("Invalid VALUE_FIELD index");
			}
			targetP->value._fieldP = g_sourceP->m_fieldsP + val;
		}
		P += value_arg;
		break;
	case VALUE_METHOD:		// 0x1a	size - 1 (03) ubyte[size] unsigned (zero-extended) four-byte integer value, interpreted as an index into the method_ids section and representing a reflective method value  
		if (value_arg > 3) {
			Cdexerror::error("Invalid VALUE_METHOD");
		}
		++value_arg;
		if (load) {
			u4T	val;

			val = decode_u4(value_arg, P);
			if (g_sourceP->m_methods <= val) {
				Cdexerror::error("Invalid VALUE_METHOD index");
			}
			targetP->value._methodP = g_sourceP->m_methodsP + val;
		}
		P += value_arg;
		break;	
	case VALUE_ENUM:		// 0x1b	size - 1 (03) ubyte[size] unsigned (zero-extended) four-byte integer value, interpreted as an index into the field_ids section and representing the value of an enumerated type constant  
		if (value_arg > 3) {
			Cdexerror::error("Invalid VALUE_ENUM");
		}
		++value_arg;
		if (load) {
			u4T	val;
			
			val = decode_u4(value_arg, P);
			if (g_sourceP->m_fields <= val) {
				Cdexerror::error("Invalid VALUE_ENUM index");
			}
			targetP->value._fieldP = g_sourceP->m_fieldsP + val;
		}
		P += value_arg;
		break;	
	case VALUE_ARRAY:		// 0x1c	(none; must be 0) encoded_array an array of values, in the format specified by "encoded_array Format" below. The size of the value is implicit in the encoding.  

		if (value_arg) {
			Cdexerror::error("Invalid VALUE_ARRAY");
		}
		encoded_array(&P, &targetP->value._arrayP, memoryPP, load);
		break;
	case VALUE_ANNOTATION:	// 0x1d	(none; must be 0) encoded_annotation a sub-annotation, in the format specified by "encoded_annotation Format" below. The size of the value is implicit in the encoding.  
	{	
		annotationT	*annotationP;

		if (value_arg) {
			Cdexerror::error("Invalid VALUE_ANNOTATION");
		}
		annotationP = (annotationT *) *memoryPP;
		if (load) {
			targetP->value._annotationP = annotationP;
		}
		*memoryPP = (u1T *) (annotationP + 1);
		Cencoded_annotation::scan(&P, annotationP, memoryPP, load);
		break;
	}
	case VALUE_NULL:		// 0x1e	(none; must be 0) (none) null reference value 
		if (value_arg) {
			Cdexerror::error("Invalid VALUE_NULL");
		}
		break;
	case VALUE_BOOLEAN:		// 0x1f	boolean (01) (none) one-bit value; 0 for false and 1 for true. The bit is represented in the value_arg.
		if (value_arg > 1) {
			Cdexerror::error("Invalid VALUE_BOOLEAN");
		}
		if (load) {
			targetP->value._s4 = value_arg;
		}
		break;
	default:
		Cdexerror::error("Unknown value_type");
	}
	*PP = P;
}

/* static */ void
Cencoded_annotation::scan(u1T **PP, annotationT *targetP, u1T **memoryPP, int load)
{
	u1T					*P;
	u4T					type_idx;
	u4T					size;
	annotation_elementT	*elementP;
	annotation_valueT	*valueP;
	int					i;
	u4T					name_idx;

	P        = *PP;
	type_idx = decode_uleb128(&P);
	size     = decode_uleb128(&P);
	elementP = ((annotation_elementT *) *memoryPP);
	if (load) {
		targetP->typeP     = getType(type_idx);
		targetP->size      = size;
		targetP->elementsP = elementP; 
	}
	*memoryPP = (u1T *) (elementP + size);

	for (i = size; i; --i) {
		// nnotation_element
		name_idx  = decode_uleb128(&P);
		valueP    = (annotation_valueT *) *memoryPP;
		*memoryPP = (u1T *) (valueP+1);
		if (load) {
			elementP->nameP  = getString(name_idx);
			elementP->valueP = valueP;
			++elementP;
		}
		Cencoded_value::scan(&P, valueP, memoryPP, load);
}	}	

/* static */ void
Cannotation_item::scan(u4T offset, annotation_itemT *targetP, u1T **memoryPP, int load)
{
	u1T					*P;

	in_data_section(offset);
	P       = g_imageP + offset;
	if (load) {
		targetP->visibility = *P;
	}
	++P;
	Cencoded_annotation::scan(&P, &targetP->annotation, memoryPP, load);
}

void
Cannotation_off_item::scan(annotation_itemT *targetP, u1T **memoryPP, int load)
{
	Cannotation_item::scan(annotation_off, targetP, memoryPP, load);
}

/* static */ void
Cannotation_set_item::check_offset(u4T offset)
{
	in_data_section(offset);
	if (offset & 3) {
		Cdexerror::error("annotation_set_item.offset unaligned");
}	}

/* static */ void
Cannotation_set_item::fixup(u4T offset)
{
	Cannotation_set_item	*annotation_set_itemP;

	check_offset(offset);
	annotation_set_itemP = (Cannotation_set_item *) (g_imageP + offset);
	fixup_uint((u1T *) &annotation_set_itemP->size);
}

/* static */ void
Cannotation_set_item::scan(u4T offset, annotation_setT *targetP, u1T **memoryPP, int load)
{
	Cannotation_set_item	*annotation_set_itemP;
	Cannotation_off_item	*entriesP, *entryP, *endP;
	annotation_itemT		*itemP;
	u4T						size;

	check_offset(offset);
	annotation_set_itemP = (Cannotation_set_item *) (g_imageP + offset);
	size                 = annotation_set_itemP->size;
	itemP                = (annotation_itemT *) *memoryPP;
	*memoryPP            = (u1T *) (itemP + size);
	if (load) {
		targetP->size    = size;
		targetP->itemsP  = itemP;
	}

	entriesP = entryP    = annotation_set_itemP->entries;
	for (endP = entriesP + size; entryP < endP; ++entryP) {
		entryP->scan(itemP, memoryPP, load);
		++itemP;
}	}

void
Cannotation_set_ref_item::fixup(void)
{
	u4T						offset;

	if ((offset = annotations_off)) {
		Cannotation_set_item::fixup(offset);
}	}

/* static */ void
Cannotation_set_ref_list::check_offset(u4T offset)
{
	in_data_section(offset);
	if (offset & 3) {
		Cdexerror::error("Unaligned annotation_set_ref_list");
}	}

/* static */ void
Cannotation_set_ref_list::fixup(u4T offset)
{
	Cannotation_set_ref_list	*annotation_set_ref_listP;
	Cannotation_set_ref_item	*listP;
	u4T	i;

	check_offset(offset);
	annotation_set_ref_listP = (Cannotation_set_ref_list *) (g_imageP + offset);
	listP                    = annotation_set_ref_listP->list;
	fixup_uint((u1T *) &annotation_set_ref_listP->size);
	for (i = annotation_set_ref_listP->size; i; --i) {
		listP->fixup();
		++listP;
}	}

void
Cfield_annotation::fixup(void)
{

	fixup_uint((u1T *) &field_idx);
	fixup_uint((u1T *) &annotations_off);
	Cannotation_set_item::fixup(annotations_off);
}

void
Cfield_annotation::scan(annotation_setT *targetP, u1T **memoryPP, int load)
{

	if (g_headerP->field_ids_size <= field_idx) {
		Cdexerror::error("Invalid field_annotation.field_idx");
	}
	Cannotation_set_item::scan(annotations_off, targetP, memoryPP, load);
}

void
Cmethod_annotation::fixup(void)
{
	fixup_uint((u1T *) &method_idx);
	fixup_uint((u1T *) &annotations_off);
	Cannotation_set_item::fixup(annotations_off);
}

void
Cmethod_annotation::scan(annotation_setT *targetP, u1T **memoryPP, int load)
{
	if (g_headerP->method_ids_size <= method_idx) {
		Cdexerror::error("Invalid method_annotation.method_idx");
	}
	Cannotation_set_item::scan(annotations_off, targetP, memoryPP, load);
}

void
Cparameter_annotation::fixup(void)
{
	fixup_uint((u1T *) &method_idx);
	fixup_uint((u1T *) &annotations_off);
	Cannotation_set_ref_list::fixup(annotations_off);
};

void
Cparameter_annotation::scan(annotation_parameterT **targetPP, u1T **memoryPP, int load)
{
	u4T							offset;
	Cannotation_set_ref_list	*annotation_set_ref_listP;
	Cannotation_set_ref_item	*listP;
	u4T							size;
	annotation_parameterT		*targetP;
	annotation_setT				*setP, *end_setP;

	if (g_sourceP->m_methods <= method_idx) {
		Cdexerror::error("Invalid parameter_annotation.method_idx");
	}

	offset                   = annotations_off;
	Cannotation_set_ref_list::check_offset(offset);
	annotation_set_ref_listP = (Cannotation_set_ref_list *) (g_imageP + offset);
	listP                    = annotation_set_ref_listP->list;
	size                     = annotation_set_ref_listP->size;
	targetP                  = (annotation_parameterT *) *memoryPP;
	*memoryPP                = (u1T *) (targetP->list + size);

	if (load) {
		targetP->method = method_idx;
		targetP->size   = size;
		*targetPP       = targetP;
	}

	setP     = targetP->list;
	for (end_setP = setP + size; setP < end_setP; ++setP) {
		offset = listP->annotations_off;
		if (offset) {
			Cannotation_set_item::scan(offset, setP, memoryPP, load);
		} else {
			setP->size   = 0;
			setP->itemsP = 0;
}	}	}

/* static */ void
Cannotations_directory_item::check_offset(u4T offset)
{
	in_data_section(offset);
	if (offset & 3) {
		Cdexerror::error("Annotations_directory_item not aligned");
}	}

/* static */ void
Cannotations_directory_item::fixup(u4T offset)
{
	Cannotations_directory_item	*annotations_directory_itemP;
	Cfield_annotation			*field_annotationP;
	Cmethod_annotation			*method_annotationP;
	Cparameter_annotation		*parameter_annotationP;
	u4T	i;

	check_offset(offset);
	annotations_directory_itemP = (Cannotations_directory_item *) (g_imageP + offset);
	fixup_uint((u1T *) &annotations_directory_itemP->class_annotations_off);
	if (annotations_directory_itemP->class_annotations_off) {
		Cannotation_set_item::fixup(annotations_directory_itemP->class_annotations_off);
	}
	fixup_uint((u1T *) &annotations_directory_itemP->fields_size);
	field_annotationP = annotations_directory_itemP->field_annotations;
	for (i = annotations_directory_itemP->fields_size; i; --i) {
		field_annotationP->fixup();
		++field_annotationP;
	}

	fixup_uint((u1T *) &annotations_directory_itemP->methods_size);
	method_annotationP = (Cmethod_annotation *) field_annotationP;
	for (i = annotations_directory_itemP->methods_size; i; --i) {
		method_annotationP->fixup();
		++method_annotationP;
	}

	fixup_uint((u1T *) &annotations_directory_itemP->parameters_size);
	parameter_annotationP = (Cparameter_annotation *) method_annotationP;
	for (i = annotations_directory_itemP->parameters_size; i; --i) {
		parameter_annotationP->fixup();
		++parameter_annotationP;
}	}

/* static */ void
Cannotations_directory_item::scan(annotationsT *targetP, u4T offset)
{
	Cannotations_directory_item	*annotations_directory_itemP;
	Cfield_annotation			*fieldsP, *fieldP, *end_fieldsP;
	u4T							fields;
	Cmethod_annotation			*methodsP, *methodP, *end_methodsP;
	u4T							methods;
	Cparameter_annotation		*parametersP, *parameterP, *end_parametersP;
	u4T							parameters;
	u1T							*memoryP;
	annotation_setT				*setP;
	annotation_parameterT		**parametersPP;
	int							load;
	u4T							offset1, need;

	check_offset(offset);
	annotations_directory_itemP = (Cannotations_directory_item *) (g_imageP + offset);
	fields                      = annotations_directory_itemP->fields_size;
	methods                     = annotations_directory_itemP->methods_size;
	parameters                  = annotations_directory_itemP->parameters_size;

	fieldsP                     = annotations_directory_itemP->field_annotations;
	end_fieldsP                 = fieldsP + fields;

	methodsP                    = (Cmethod_annotation *) end_fieldsP;
	end_methodsP                = methodsP + methods;

	parametersP                 = (Cparameter_annotation *) end_methodsP;
	end_parametersP             = parametersP + parameters;

	memoryP  = 0;
	memoryP += (fields + methods + parameters) * sizeof(annotation_setT);
	for (load = 0; ; ++load) { 
		if ((offset1 = annotations_directory_itemP->class_annotations_off)) {
			Cannotation_set_item::scan(offset1, &targetP->class_annotations, &memoryP, load);
		}
	
		setP  = targetP->fieldsP;
		for (fieldP = fieldsP; fieldP < end_fieldsP; ++fieldP) {
			fieldP->scan(setP, &memoryP, load);
			++setP;
		}

		setP = targetP->methodsP;
		for (methodP = methodsP; methodP < end_methodsP; ++methodP) {
			methodP->scan(setP, &memoryP, load);
			++setP;
		}

		parametersPP = targetP->parametersPP;
		for (parameterP = parametersP; parameterP < end_parametersP; ++parameterP) {
			parameterP->scan(parametersPP, &memoryP, load);
			++parametersPP;
		}
		if (load) {
			break;
		}
		need                  = memoryP - (u1T *) 0;
		targetP->fieldsP      = (annotation_setT *) Xmalloc(need);
		targetP->methodsP     = targetP->fieldsP  + fields;
		targetP->parametersPP = (annotation_parameterT **) (targetP->methodsP + methods);
		memoryP               = (u1T *) (targetP->parametersPP + parameters);
}	}

// =================================================================================

/* static */ void
Cencoded_field::fixup(u1T **PP)
{
	u4T field_idx_diff;
	u4T access_flags;

	field_idx_diff = decode_uleb128(PP);
	access_flags   = decode_uleb128(PP);
}

void
Ctry_item::fixup(void)
{
	fixup_uint( (u1T *) &start_addr);
	fixup_ushort((u1T *) &insn_count);
	fixup_ushort((u1T *) &handler_off);
}

/* static */ void
Ccode_item::check_offset(u4T offset)
{
	in_data_section(offset);
	if (offset & 3) {
		Cdexerror::error("code_item unaligned");
}	}

/* static */ void
Ccode_item::fixup(u4T offset)
{
	Ccode_item  *code_itemP;
	u2T			*start_codeP;
	Ctry_item	*try_itemP;
	int			i;

	check_offset(offset);
	code_itemP = (Ccode_item *) (g_imageP + offset);
	fixup_ushort((u1T *) &code_itemP->registers_size);
	fixup_ushort((u1T *) &code_itemP->ins_size);
	fixup_ushort((u1T *) &code_itemP->outs_size);
	fixup_ushort((u1T *) &code_itemP->tries_size);
	fixup_uint((u1T *)	&code_itemP->debug_info_off);
	fixup_uint((u1T *)	&code_itemP->insns_size);

	start_codeP = code_itemP->insns;
	try_itemP   = (Ctry_item *) (start_codeP + ((code_itemP->insns_size + 1) & ~1));

	for (i = code_itemP->tries_size; i; --i) {
		try_itemP->fixup();
		++try_itemP;
}	}

/* static */ void
Cdebug_info_item::scan(u4T offset)
{
	Cfile		*source_fileP;
	u1T			*debug_infoP, *P;
	u4T			line_start, parameters;
	u4T			parameter;
	
	u4T			load;
	u1T			pcode;

	u4T			variable_entries;
	variableT	*variablesP, *variableP, variable;
	u4T			position_entries;
	positionT	*positionP, position;
	u4T			i;

	g_functionP->m_variables  = 0;
	g_functionP->m_variablesP = 0;
	g_functionP->m_positions  = 0;
	g_functionP->m_positionsP = 0;
	if (!offset) {
		return;
	}
	in_data_section(offset);
	source_fileP                = 0;
	debug_infoP                 = g_imageP + offset;
	line_start                  = decode_uleb128(&debug_infoP);
	g_functionP->m_parameters   = parameters = decode_uleb128(&debug_infoP);
	g_functionP->m_parametersPP = 0;

	if (parameters) {
		g_functionP->m_parametersPP = (stringT **) Xmalloc(sizeof(stringT) * parameters);
		for (i = 0; i < parameters; ++i) {
			parameter = decode_uleb128(&debug_infoP);
			if (!parameter) {
				g_functionP->m_parametersPP[i] = 0;
			} else {
				g_functionP->m_parametersPP[i] = getString(parameter - 1);
	}	}	}

	for (load = 0; ; ++load) {
		position.address	    = 0;			// Must monitonically increase
		position.line		    = line_start;	// Must be >= 1
		position.source_fileP   = g_functionP->m_classP->m_source_fileP;
		position.flags          = 0;

		variable_entries = 0;
		position_entries = 0;
		for (P = debug_infoP;;) {
			switch (pcode = *P++) {
			case 0x00:	// DBG_END_SEQUENCE (none) terminates a debug info sequence for a code_item 
				break;
			case 0x01:	// DBG_ADVANCE_PC uleb128 addr_diff addr_diff: amount to add to address register advances the address register without emitting a positions entry
				position.address += decode_uleb128(&P);
				continue;
			case 0x02:	// DBG_ADVANCE_LINE sleb128 line_diff line_diff: amount to change line register by advances the line register without emitting a positions entry 
				position.line    += decode_sleb128(&P);
				assert(position.line >= 1);
				continue;
			case 0x03:	// DBG_START_LOCAL uleb128 register_num uleb128p1 name_idx uleb128p1 type_idx 
			{
				u4T	name_idx, type_idx;

				variable.register_num = decode_uleb128(&P);
				name_idx              = decode_uleb128(&P) - 1;
				type_idx              = decode_uleb128(&P) - 1;

				if (load) {
					if (name_idx == NO_INDEX) {
						variable.nameP = 0;
					} else {
						variable.nameP = getString(name_idx);
					}
					variable.typeP         = getType(type_idx);
					variable.sigP          = 0;
					variable.start_address = position.address;
					variable.end_address   = 0;
					*variableP++           = variable;
				}
				++variable_entries;
				continue;
			}
			case 0x04:	// DBG_START_LOCAL_EXTENDED 0x04 uleb128 register_num uleb128p1 name_idx uleb128p1 type_idx uleb128p1 sig_idx
			{
				u4T	name_idx, type_idx, sig_idx;

				variable.register_num = decode_uleb128(&P);
				name_idx              = decode_uleb128(&P) - 1;
				type_idx              = decode_uleb128(&P) - 1;
				sig_idx               = decode_uleb128(&P) - 1;
				if (load) {
					variable.nameP         = getString(name_idx);
					variable.typeP         = getType(type_idx);
					variable.sigP          = getType(sig_idx);
					variable.start_address = position.address;
					variable.end_address   = 0;
					*variableP++           = variable;
				}
				++variable_entries;
				continue;
			}
			case 0x05:	// DBG_END_LOCAL uleb128 register_num register_num: register that contained local marks a currently-live local variable as out of scope at the current address  
				variable.register_num = decode_uleb128(&P);
				if (load) {
					for (i = variable_entries; --i >= 0; ) {
						if (variablesP[i].register_num == variable.register_num && !variablesP[i].end_address) {
							variablesP[i].end_address = position.address;
							break;
				}	}	}
				continue;
			case 0x06:	// DBG_RESTART_LOCAL uleb128 register_num register_num: register to restart re-introduces a local variable at the current address. 
						// The name and type are the same as the last local that was live in the specified register.  
				variable.register_num = decode_uleb128(&P);
				if (load) {
					for (i = variable_entries; ; ) {
						if (--i < 0) {
							variable.nameP     = 0;
							variable.typeP     = 0;
							variable.sigP      = 0;
							break;
						}
						if (variablesP[i].register_num == variable.register_num) {
							variable = variablesP[i];
							break;
					}	}
					variable.start_address = position.address;
					variable.end_address   = 0;
					*variableP++ = variable;
				}
				++variable_entries;
				continue;
			case 0x07:	// DBG_SET_PROLOGUE_END (none) sets the prologue_end state machine register, indicating that the next position entry that is added should be considered the end of a method prologue (an appropriate place for a method breakpoint). The prologue_end register is cleared by any special (>= 0x0a) opcode.  
				position.flags |= Prologue_endX;
				continue;
			case 0x08:	// DBG_SET_EPILOGUE_BEGIN (none) sets the epilogue_begin state machine register, indicating that the next position entry that is added should be considered the beginning of a method epilogue (an appropriate place to suspend execution before method exit). The epilogue_begin register is cleared by any special (>= 0x0a) opcode.  
				position.flags |= Epilogue_beginX;
				continue;
			case 0x09:	// DBG_SET_FILE uleb128p1 name_idx name_idx: string index of source file name; NO_INDEX if unknown  indicates that all subsequent line number entries make reference to this source file name, instead of the default name specified in code_item  
				i = decode_uleb128(&P);
				if (!i) {
					source_fileP = 0;
				} else {
					source_fileP = Cfile::locate(getString(i - 1));
				}
				position.source_fileP = source_fileP;	
				continue;
			default:	// Special Opcodes 0x0a0xff
				pcode            -= 0x0a;
				position.line    += (pcode % 15) - 4;
				assert(position.line >= 1);
				position.address += (pcode / 15);
				if (load) {
					*positionP++ = position;
				}
				position.flags = 0;
				++position_entries;
				continue;
			}
			break;
		}

		if (load) {
			break;
		}

		g_functionP->m_variables  = variable_entries;
		if (variable_entries) {
			variablesP = (variableT *) Xmalloc(variable_entries * sizeof(variableT));
		} else {
			variablesP = 0;
		}
		g_functionP->m_variablesP = variableP = variablesP;
		g_functionP->m_positions  = position_entries;
		if (position_entries) {
			positionP = (positionT *) Xmalloc(position_entries * sizeof(positionT));
		} else {
			positionP = 0;
		}
		g_functionP->m_positionsP = positionP;
	}
}

static u4T
getU4(u2T **dataPP)
{
	u2T	*dataP;
	u4T	high, ret;

	dataP = *dataPP;
	ret   = *dataP++;
	high  = *dataP++;
	ret  |= (high << 16);
	*dataPP = dataP;
	return(ret);
}

static u8T
getU8(u2T **dataPP)
{
	u2T	*dataP;
	u8T	high, ret;

	dataP = *dataPP;
	ret   = *dataP++;
	high  = *dataP++;
	ret  |= (high << 16);
	high  = *dataP++;
	ret  |= (high << 32);
	high  = *dataP++;
	ret  |= (high << 48);
	*dataPP = dataP;
	return(ret);
}

u4T	g_pc;

static void
set_register(registerT *registerP, u4T index)
{
	variableT	*variableP, *end_variablesP;

	registerP->number = index;
	variableP         = g_functionP->m_variablesP;
	for (end_variablesP = variableP + g_functionP->m_variables; variableP < end_variablesP; ++variableP) {
		if (variableP->register_num == index) {
			if (g_pc >= variableP->start_address && g_pc < variableP->end_address) {
				registerP->u.variableP = variableP;
				return;
	}	}	}
	registerP->u.variableP = 0;
}

static void
set_string(registerT *registerP, u4T index)
{
	registerP->number = index;
	registerP->u.stringP = getString(index);
}

static void
set_type(registerT *registerP, u4T index)
{
	registerP->number = index;
	registerP->u.typeP   = getType(index);
}

static void
set_field(registerT *registerP, u4T index)
{
	registerP->number = index;
	registerP->u.fieldP = getField(index);
}

static void
set_method(registerT *registerP, u4T index)
{
	registerP->number = index;
	registerP->u.methodP = getMethod(index);
}

static void
fixup_branches(codeT *startP, codeT *endP)
{
	extern int	g_match_lines;

	codeT		*assemblerP, *assembler1P;
	s4T			branch;
	u4T			pc;
	positionT	*position1P, *position2P;

	for (assemblerP = startP; assemblerP < endP; ++assemblerP) {
		switch (assemblerP->m_pcode) {
		case GOTO_8:			// 0x28: 10t ->  AA|op -> op +AA
		case GOTO_16:			// 0x29: 20t -> |op AAAA -> op +AAAA
		case GOTO_32:			// 0x2a:		// 30t -> |op AAAAlo AAAAhi -> op +AAAAAAAA
		case IF_EQ_ZERO:		// 0x38: 21t -> AA|op BBBB -> op vAA, +BBBB
		case IF_NE_ZERO:		// 0x39:
		case IF_LT_ZERO:		// 0x3a:
		case IF_GE_ZERO:		// 0x3b:
		case IF_GT_ZERO:		// 0x3c:
		case IF_LE_ZERO:		// 0x3d:
		case IF_EQ:				// 0x32: 22t -> B|A|op CCCC -> op vA, vB, +CCCC
		case IF_NE:				// 0x33:
		case IF_LT:				// 0x34:
		case IF_GE:				// 0x35:
		case IF_GT:				// 0x36:
		case IF_LE:				// 0x37:
		case PACKED_SWITCH:		// 0x2b: 31t -> AA|op BBBBlo BBBBhi -> op vAA, +BBBBBBBB (with supplemental data as specified below in "packed-switch Format")
		case SPARSE_SWITCH:		// 0x2c: 31t -> AA|op BBBBlo BBBBhi -> op vAA, +BBBBBBBB  (with supplemental data as specified below in "sparse-switch Format")
		case TRY_START:			//
			branch = assemblerP->m_jump.offset;
			pc     = assemblerP->m_pc + branch;
			if (branch < 0) {
				for (assembler1P = assemblerP;; ) {
					if (assembler1P == startP) {
						assert(0);
						assemblerP->m_jump.P = 0;
						break;
					}
					--assembler1P;
					if (assembler1P->m_pc == pc) {
						assemblerP->m_jump.P = assembler1P;
						break;
				}	}
			} else {
				for (assembler1P = assemblerP; ; ++ assembler1P) {
					if (assembler1P == endP) {
						assert(0);
						assemblerP->m_jump.P = 0;
						break;
					}
					if (assembler1P->m_pc == pc) {
						assemblerP->m_jump.P = assembler1P;
						break;
			}	}	}
			break;	
		default:
			assemblerP->m_jump.P = 0;
		}
		if (g_match_lines) {
			if (assemblerP != startP) {
				position1P = assemblerP[0].m_positionP;
				if (position1P) {
					position2P = assemblerP[-1].m_positionP;
					if (position2P) {
						if (position1P->line == position2P->line && position1P->source_fileP == position2P->source_fileP) {
							// Don't hash if not first instruction on a line
							continue;
		}	}	}	}	}
		hash_pcode(assemblerP);
}	}

static void
load_code(u2T *start_codeP, u4T size, codeT **assemblerPP, u1T **extraPP, int load)
{
	u2T			*codeP;
	u2T			*end_codeP;
	codeT		*startP, *assemblerP;
	u1T			*extraP;
	labelT		*labelP, *end_labelsP;
	positionT	*positionP, *end_positionP;
	u2T			code;
	pcodeE		pcode;
	u4T			width, entries;

	positionP     = g_functionP->m_positionsP;
	end_positionP = positionP + g_functionP->m_positions - 1;

	assemblerP  = startP = *assemblerPP;
	extraP      = *extraPP;
	labelP      = Csort::g_labelsP;
	end_labelsP = labelP + Csort::g_labels_count;

	end_codeP  = start_codeP + size;
	for (codeP = start_codeP; codeP < end_codeP; ) {
		code    = *codeP;
		switch (code) {
		case 0x300:			// Fill array data
			++codeP;
			width   = *codeP++;
			entries = getU4(&codeP);
			codeP  += (entries * width + 1) >> 1;
			assert(codeP <= end_codeP);
			continue;
		case 0x200:			// Sparse switch data
			++codeP;
			codeP += (*codeP << 2) + 1;
			assert(codeP <= end_codeP);
			continue;
		case 0x100:			// Packed switch data
			++codeP;
			codeP += (*codeP << 1) + 3;
			assert(codeP <= end_codeP);
			continue;
		}
		g_pc    = codeP - start_codeP;
		pcode   = (pcodeE) (code & 0xFF);

		if (load) {
			assert((assemblerP - g_functionP->m_assemblersP) < (int) g_functionP->m_assemblers);
			assemblerP->m_pc        = g_pc;
			assemblerP->m_pcode     = pcode;
			assemblerP->m_functionP = g_functionP;
			assemblerP->m_matchP    = 0;
			assemblerP->m_hashP     = 0;
			assemblerP->info.P      = extraP;
		
			for (; positionP < end_positionP && positionP[1].address <= g_pc; ++positionP);
			if (positionP <= end_positionP) {
				assemblerP->m_positionP = positionP;
			} else {
				assemblerP->m_positionP = 0;
			}	
		}
		if (labelP < end_labelsP) {
			if (labelP->from <= g_pc) {
				assert(labelP->from == g_pc);
				if (load) {
					if (labelP->enabled) {
						assemblerP->m_pcode = TRY_START;
					} else {
						assemblerP->m_pcode = TRY_END;
					}
					((info_tryT *) extraP)->to        = labelP->to;
					set_type(&((info_tryT *) extraP)->classtype, labelP->classtype);
					assemblerP->m_jump.offset = labelP->handler - g_pc;
				}
				extraP += sizeof(info_tryT);
				++assemblerP;
				++labelP;
				continue;
		}	}

		++codeP;
		switch (pcode) {
		// No registers
		case NOP:				// 0x00: 10x ->  00|op  -> op
		case RETURN_VOID:		// 0x0e:
			if (code & 0xFF00) {
				Cdexerror::error("Illegal non-zero filler in code");
			}
			if (load) {
				assemblerP->info.P = 0;
			}
			break;
		// One register
		case MOVE_RESULT:		// 0x0a: 11x ->  AA|op  -> op vAA 
		case MOVE_RESULT_WIDE:	// 0x0b
		case MOVE_RESULT_OBJECT:// 0x0c
		case MOVE_EXCEPTION:	// 0x0d
		case RETURN:			// 0x0f
		case RETURN_WIDE:		// 0x10
		case RETURN_OBJECT:		// 0x11
		case MONITOR_ENTER:		// 0x1d
		case MONITOR_EXIT:		// 0x1e
		case THROW:				// 0x27
			if (load) {
				set_register(&((info_register1T *) extraP)->reg1, code >> 8);
			}
			extraP += sizeof(info_register1T);
			break;
		// Two registers
		case ARRAY_LENGTH:		// 0x21: 12x -> B|A|op -> op vA, vB 
		case NEG_INT:			// 0x7b:
		case NOT_INT:			// 0x7c:
		case NEG_LONG:			// 0x7d:
		case NOT_LONG:			// 0x7e:
		case NEG_FLOAT:			// 0x7f:
		case NEG_DOUBLE:		// 0x80:
		case INT_TO_LONG:		// 0x81:
		case INT_TO_FLOAT:		// 0x82:
		case INT_TO_DOUBLE:		// 0x83:
		case LONG_TO_INT:		// 0x84:
		case LONG_TO_FLOAT:		// 0x85:
		case LONG_TO_DOUBLE:	// 0x86:
		case FLOAT_TO_INT:		// 0x87:
		case FLOAT_TO_LONG:		// 0x88:
		case FLOAT_TO_DOUBLE:	// 0x89:
		case DOUBLE_TO_INT:		// 0x8a:
		case DOUBLE_TO_LONG:	// 0x8b:
		case DOUBLE_TO_FLOAT:	// 0x8c:
		case INT_TO_BYTE:		// 0x8d:
		case INT_TO_CHAR:		// 0x8e:
		case INT_TO_SHORT:		// 0x8f:
		case ADD_INT2:			// 0xb0:
		case SUB_INT2:			// 0xb1:
		case MUL_INT2:			// 0xb2:
		case DIV_INT2:			// 0xb3:
		case REM_INT2:			// 0xb4:
		case AND_INT2:			// 0xb5:
		case OR_INT2:			// 0xb6:
		case XOR_INT2:			// 0xb7:
		case SHIFT_LEFT_INT2:	// 0xb8:
		case SHIFT_RIGHT_INT2:	// 0xb9:
		case USHIFT_RIGHT_INT2:	// 0xba:
		case ADD_LONG2:			// 0xbb:
		case SUB_LONG2:			// 0xbc:
		case MUL_LONG2:			// 0xbd:
		case DIV_LONG2:			// 0xbe:
		case REM_LONG2:			// 0xbf:
		case AND_LONG2:			// 0xc0:
		case OR_LONG2:			// 0xc1:
		case XOR_LONG2:			// 0xc2:
		case SHIFT_LEFT_LONG2:	// 0xc3:
		case SHIFT_RIGHT_LONG2:	// 0xc4:
		case USHIFT_RIGHT_LONG2:// 0xc5:
		case ADD_FLOAT2:		// 0xc6:
		case SUB_FLOAT2:		// 0xc7:
		case MUL_FLOAT2:		// 0xc8:
		case DIV_FLOAT2:		// 0xc9:
		case REM_FLOAT2:		// 0xca:
		case ADD_DOUBLE2:		// 0xcb:
		case SUB_DOUBLE2:		// 0xcc:
		case MUL_DOUBLE2:		// 0xcd:
		case DIV_DOUBLE2:		// 0xce:
		case REM_DOUBLE2:		// 0xcf:
			if (load) {
				set_register(&((info_register2T *) extraP)->reg1, (code >>  8) & 0xF);	// A
				set_register(&((info_register2T *) extraP)->reg2,  code >> 12);			// B
			}
			extraP += sizeof(info_register2T);
			break;
		case MOVE_4_4:			// 0x01: 12x -> B|A|op -> op vA, vB
			if (load) {
				assemblerP->m_pcode = MOVE_16_16;	// Normalize
				set_register(&((info_register2T *) extraP)->reg1, (code >>  8) & 0xF);	// A
				set_register(&((info_register2T *) extraP)->reg2,  code >> 12);			// B
			}
			extraP += sizeof(info_register2T);
			break;
		case MOVE_WIDE_4_4:		// 0x04
			if (load) {
				assemblerP->m_pcode = MOVE_WIDE_16_16;	// Normalize
				set_register(&((info_register2T *) extraP)->reg1, (code >>  8) & 0xF);	// A
				set_register(&((info_register2T *) extraP)->reg2,  code >> 12);			// B
			}
			extraP += sizeof(info_register2T);
			break;
		case MOVE_OBJECT_4_4:	// 0x07
			if (load) {
				assemblerP->m_pcode = MOVE_OBJECT_16_16;	// Normalize
				set_register(&((info_register2T *) extraP)->reg1, (code >>  8) & 0xF);	// A
				set_register(&((info_register2T *) extraP)->reg2,  code >> 12);			// B
			}
			extraP += sizeof(info_register2T);
			break;

		case MOVE_8_16:			// 0x02: 22x -> AA|op BBBB -> op vAA, vBBBB
			if (load) {
				assemblerP->m_pcode = MOVE_16_16;	// Normalize
				set_register(&((info_register2T *) extraP)->reg1, code >>  8);			// A
				set_register(&((info_register2T *) extraP)->reg2, *codeP);				// B
			}
			++codeP;
			extraP += sizeof(info_register2T);
			break;
		case MOVE_WIDE_8_16:	// 0x05:
			if (load) {
				assemblerP->m_pcode = MOVE_WIDE_16_16;	// Normalize
				set_register(&((info_register2T *) extraP)->reg1,  code >>  8);			// A
				set_register(&((info_register2T *) extraP)->reg2,  *codeP);				// B
			}
			++codeP;
			extraP += sizeof(info_register2T);
			break;
		case MOVE_OBJECT_8_16:	// 0x08:
			if (load) {
				assemblerP->m_pcode = MOVE_OBJECT_16_16;	// Normalize
				set_register(&((info_register2T *) extraP)->reg1, code >>  8);			// A
				set_register(&((info_register2T *) extraP)->reg2, *codeP);				// B
			}
			++codeP;
			extraP += sizeof(info_register2T);
			break;
		case MOVE_16_16:		// 0x03: 32x -> |op AAAA BBBB -> op vAAAA, vBBBB
		case MOVE_WIDE_16_16:	// 0x06:
		case MOVE_OBJECT_16_16:	// 0x09:
			if (load) {
				set_register(&((info_register2T *) extraP)->reg1, codeP[0]);			// A
				set_register(&((info_register2T *) extraP)->reg2, codeP[1]);			// B
			}
			codeP  += 2;
			extraP += sizeof(info_register2T);
			break;

		// Three registers
		case CMPL_FLOAT:		// 0x2d: 23x -> AA|op CC|BB -> op vAA, vBB, vCC
		case CMPG_FLOAT:		// 0x2e:
		case CMPL_DOUBLE:		// 0x2f:
		case CMPG_DOUBLE:		// 0x30:
		case CMP_LONG:			// 0x31:

		case AGET:				// 0x44:
		case AGET_WIDE:			// 0x45:
		case AGET_OBJECT:		// 0x46:
		case AGET_BOOLEAN:		// 0x47:
		case AGET_BYTE:			// 0x48:
		case AGET_CHAR:			// 0x49:
		case AGET_SHORT:		// 0x4a:
		case APUT:				// 0x4b:
		case APUT_WIDE:			// 0x4c:
		case APUT_OBJECT:		// 0x4d:
		case APUT_BOOLEAN:		// 0x4e:
		case APUT_BYTE:			// 0x4f:
		case APUT_CHAR:			// 0x50:
		case APUT_SHORT:		// 0x51:

		case ADD_INT:			// 0x90:
		case SUB_INT:			// 0x91:
		case MUL_INT:			// 0x92:
		case DIV_INT:			// 0x93:
		case REM_INT:			// 0x94:
		case AND_INT:			// 0x95:
		case OR_INT:			// 0x96:
		case XOR_INT:			// 0x97:
		case SHIFT_LEFT_INT:	// 0x98:
		case SHIFT_RIGHT_INT:	// 0x99:
		case USHIFT_RIGHT_INT:	// 0x9a:
		case ADD_LONG:			// 0x9b:
		case SUB_LONG:			// 0x9c:
		case MUL_LONG:			// 0x9d:
		case DIV_LONG:			// 0x9e:
		case REM_LONG:			// 0x9f:
		case AND_LONG:			// 0xa0:
		case OR_LONG:			// 0xa1:
		case XOR_LONG:			// 0xa2:
		case SHIFT_LEFT_LONG:	// 0xa3:
		case SHIFT_RIGHT_LONG:	// 0xa4:
		case USHIFT_RIGHT_LONG:	// 0xa5:
		case ADD_FLOAT:			// 0xa6:
		case SUB_FLOAT:			// 0xa7:
		case MUL_FLOAT:			// 0xa8:
		case DIV_FLOAT:			// 0xa9:
		case REM_FLOAT:			// 0xaa:
		case ADD_DOUBLE:		// 0xab:
		case SUB_DOUBLE:		// 0xac:
		case MUL_DOUBLE:		// 0xad:
		case DIV_DOUBLE:		// 0xae:
		case REM_DOUBLE:		// 0xaf:
			if (load) {
				set_register(&((info_register3T *) extraP)->reg1, code >>  8);		// A
				code         = *codeP;
				set_register(&((info_register3T *) extraP)->reg2, code & 0xFF);	// B
				set_register(&((info_register3T *) extraP)->reg3, code >> 8);		// C
			}
			++codeP;
			extraP += sizeof(info_register3T);
			break;
		// One register, one 32bit value
		case CONST_4_4:			// 0x12: 11n -> B|A|op -> op vA, #+B 
			if (load) {
				assemblerP->m_pcode = CONST_8_32;	// Normalize
				set_register(&((info_valueT *) extraP)->reg1, (code >>  8) & 0xF);	// A
				((info_valueT *) extraP)->value   = code >> 12;			// B
			}
			extraP += sizeof(info_valueT);
			break;
		case CONST_8_16:		// 0x13: 21s -> AA|op BBBB -> op vAA, #+BBBB
			if (load) {
				assemblerP->m_pcode = CONST_8_32;	// Normalize
				set_register(&((info_valueT *) extraP)->reg1, code >> 8);
				((info_valueT *) extraP)->value   =  (s2T) *codeP;
			}
			++codeP;
			extraP += sizeof(info_valueT);
			break;

		case CONST_8_32:		// 0x14: 31i -> AA|op BBBBlo BBBBhi -> op vAA, #+BBBBBBBB
		{
			u4T	value;

			value = getU4(&codeP);
			if (load) {
				set_register(&((info_valueT *) extraP)->reg1, code >> 8);
				((info_valueT *) extraP)->value   = value;
			}
			extraP += sizeof(info_valueT);
			break;
		}
		case CONST_HIGH:		// 0x15: 21h -> AA|op BBBB -> op vAA, #+BBBB0000
		case CONST_WIDE_HIGH:	// 0x19: 21h -> AA|op BBBB -> op vAA, #+BBBB000000000000 

			if (load) {
				set_register(&((info_valueT *) extraP)->reg1, (code >>  8) & 0xF);	// A
				((info_valueT *) extraP)->value =  *codeP;			// B
			}
			++codeP;
			extraP += sizeof(info_valueT);
			break;
		// One register one 64 bit value
		case CONST_WIDE_8_16:	// 0x16:
			if (load) {
				assemblerP->m_pcode = CONST_WIDE_8_64;	// Normalize
				set_register(&((info_wideT *) extraP)->reg1, code >> 8);
				((info_wideT *) extraP)->value   = (s2T) *codeP;
			}
			++codeP;
			extraP += sizeof(info_wideT);
			break;
		case CONST_WIDE_8_32:	// 0x17:
		{
			u4T	value;

			value = getU4(&codeP);
			if (load) {
				assemblerP->m_pcode = CONST_WIDE_8_64;	// Normalize
				set_register(&((info_wideT *) extraP)->reg1, code >> 8);
				((info_wideT *) extraP)->value   = (s4T) value;
			}
			extraP += sizeof(info_wideT);
			break;
		}
		case CONST_WIDE_8_64:	// 0x18: 51l -> AA|op BBBBlo BBBB BBBB BBBBhi -> op vAA, #+BBBBBBBBBBBBBBBB
		{
			u8T	value;

			value = getU8(&codeP);
			if (load) {
				set_register(&((info_wideT *) extraP)->reg1, code >> 8);
				((info_wideT *) extraP)->value   = (s8T) value;
			}
			extraP += sizeof(info_wideT);
			break;
		}
		// Two registers one 32 bit value
		case ADD_INT_LIT8:		// 0xd8: 22b -> AA|op CC|BB -> op vAA, vBB, #+CC
		case REVERSE_SUB_INT_LIT8: // 0xd9:
		case MUL_INT_LIT8:		// 0xda:
		case DIV_INT_LIT8:		// 0xdb:
		case REM_INT_LIT8:		// 0xdc:
		case AND_INT_LIT8:		// 0xdd:
		case OR_INT_LIT8:		// 0xde:
		case XOR_INT_LIT8:		// 0xdf:
		case SHIFT_LEFT_INT_LIT8://0xe0:
		case SHIFT_RIGHT_INT_LIT8://0xe1:
		case USHIFT_RIGHT_INT_LIT8://0xe2:
			if (load) {
				set_register(&((info_value2T *) extraP)->reg1, code >>  8);	// A
				code         = *codeP;
				set_register(&((info_value2T *) extraP)->reg2, code & 0xFF);	// B
				((info_value2T *) extraP)->value = code >> 8;				// C
			}
			++codeP;
			extraP += sizeof(info_value2T);
			break;
		case ADD_INT_LIT16:		// 0xd0: 22s -> B|A|op CCCC -> op vA, vB, #+CCCC
		case REVERSE_SUB_INT_LIT16:	// 0xd1:
		case MUL_INT_LIT16:		// 0xd2:
		case DIV_INT_LIT16:		// 0xd3:
		case REM_INT_LIT16:		// 0xd4:
		case AND_INT_LIT16:		// 0xd5:
		case OR_INT_LIT16:		// 0xd6:
		case XOR_INT_LIT16:		// 0xd7:
			if (load) {
				set_register(&((info_value2T *) extraP)->reg1, (code >>  8) & 0xF);	// A
				set_register(&((info_value2T *) extraP)->reg2,  code >> 12);			// B
				((info_value2T *) extraP)->value = *codeP;							// C
			}
			++codeP;
			extraP += sizeof(info_value2T);
			break;


		case CONST_STRING_8_16: // 0x1a: 21c -> AA|op BBBB -> op vAA, string@BBBB
			if (load) {
				set_register(&((info_register2T *) extraP)->reg1, (code >>  8) & 0xF);	// A
				set_string(&((info_register2T *) extraP)->reg2, *codeP);					// B
			}
			++codeP;
			extraP += sizeof(info_register2T);
			break;
		case CONST_STRING_8_32:	// 0x1b: 31c -> AA|op BBBBlo BBBBhi -> op vAA, string@BBBBBBBB
		{
			u4T	value;

			value  = getU4(&codeP);
			if (load) {
				set_register(&((info_register2T *) extraP)->reg1, code >> 8);
				set_string(  &((info_register2T *) extraP)->reg2, value);
			}
			extraP += sizeof(info_register2T);
			break;
		}

		case CONST_CLASS:		// 0x1c: 21c -> AA|op BBBB -> op vAA, type@BBBB
		case CHECK_CAST:		// 0x1f:
		case NEW_INSTANCE:		// 0x22:
			if (load) {
				set_register(&((info_register2T *) extraP)->reg1, (code >>  8) & 0xF);	// A
				set_type(&((info_register2T *) extraP)->reg2, *codeP);					// B
			}
			++codeP;
			extraP += sizeof(info_register2T);
			break;

		case STATIC_GET:		// 0x60: 21c -> AA|op BBBB -> op vAA, field@BBBB
		case STATIC_GET_WIDE:	// 0x61:
		case STATIC_GET_OBJECT:	// 0x62:
		case STATIC_GET_BOOLEAN:// 0x63:
		case STATIC_GET_BYTE:	// 0x64:
		case STATIC_GET_CHAR:	// 0x65:
		case STATIC_GET_SHORT:	// 0x66:
		case STATIC_PUT:		// 0x67:
		case STATIC_PUT_WIDE:	// 0x68:
		case STATIC_PUT_OBJECT:	// 0x69:
		case STATIC_PUT_BOOLEAN:// 0x6a:
		case STATIC_PUT_BYTE:	// 0x6b:
		case STATIC_PUT_CHAR:	// 0x6c:
		case STATIC_PUT_SHORT:	// 0x6d:
			if (load) {
				set_register(&((info_register2T *) extraP)->reg1, (code >>  8) & 0xF);	// A
				set_field(&((info_register2T *) extraP)->reg2, *codeP);					// B
			}
			++codeP;
			extraP += sizeof(info_register2T);
			break;
		case INSTANCE_OF:		// 0x20: 22c -> B|A|op CCCC -> op vA, vB, type@CCCC
		case NEW_ARRAY:			// 0x23:
			if (load) {
				set_register(&((info_register3T *) extraP)->reg1, (code >>  8) & 0xF);	// A
				set_register(&((info_register3T *) extraP)->reg2,  code >> 12);			// B
				set_type(    &((info_register3T *) extraP)->reg3, *codeP);				// C
			}
			++codeP;
			extraP += sizeof(info_register3T);
			break;
		case IGET:				// 0x52: 22c -> B|A|op CCCC -> op vA, vB, field@CCCC
		case IGET_WIDE:			// 0x53:
		case IGET_OBJECT:		// 0x54:
		case IGET_BOOLEAN:		// 0x55:
		case IGET_BYTE:			// 0x56:
		case IGET_CHAR:			// 0x57:
		case IGET_SHORT:		// 0x58:
		case IPUT:				// 0x59:
		case IPUT_WIDE:			// 0x5a:
		case IPUT_OBJECT:		// 0x5b:
		case IPUT_BOOLEAN:		// 0x5c:
		case IPUT_BYTE:			// 0x5d:
		case IPUT_CHAR:			// 0x5e:
		case IPUT_SHORT:		// 0x5f:
			if (load) {
				set_register(&((info_register3T *) extraP)->reg1, (code >>  8) & 0xF);	// A
				set_register(&((info_register3T *) extraP)->reg2,  code >> 12);			// B
				set_field(   &((info_register3T *) extraP)->reg3, *codeP);				// C
			}
			++codeP;
			extraP += sizeof(info_register3T);
			break;
		case GOTO_8:			// 0x28: 10t ->  AA|op -> op +AA
			if (load) {
				assemblerP->m_pcode = GOTO_32;		// Normalize
				assemblerP->info.P  = 0;
				assemblerP->m_jump.offset = ((s2T) code) >> 8;
			}
			break;
		case GOTO_16:			// 0x29: 20t -> |op AAAA -> op +AAAA
			if (code & 0xFF00) {
				Cdexerror::error("Illegal non-zero filler in code");
			}
			if (load) {
				assemblerP->m_pcode = GOTO_32;		// Normalize
				assemblerP->info.P  = 0;
				assemblerP->m_jump.offset = ((s2T) *codeP);
			}
			++codeP;
			break;
		case GOTO_32:			// 0x2a:		// 30t -> |op AAAAlo AAAAhi -> op +AAAAAAAA
		{
			u4T	branch;

			if (code & 0xFF00) {
				Cdexerror::error("Illegal non-zero filler in code");
			}
			branch = (s4T) getU4(&codeP);
			if (load) {
				assemblerP->info.P = 0;
				assemblerP->m_jump.offset = (s4T) branch;
			}
			break;
		}
		case IF_EQ_ZERO:		// 0x38: 21t -> AA|op BBBB -> op vAA, +BBBB
		case IF_NE_ZERO:		// 0x39:
		case IF_LT_ZERO:		// 0x3a:
		case IF_GE_ZERO:		// 0x3b:
		case IF_GT_ZERO:		// 0x3c:
		case IF_LE_ZERO:		// 0x3d:
			if (load) {
				set_register(&((info_register1T *) extraP)->reg1, code >> 8);
				assemblerP->m_jump.offset = (s2T) *codeP;
			}
			++codeP;
			extraP += sizeof(info_register1T);
			break;		
		
		case IF_EQ:				// 0x32: 22t -> B|A|op CCCC -> op vA, vB, +CCCC
		case IF_NE:				// 0x33:
		case IF_LT:				// 0x34:
		case IF_GE:				// 0x35:
		case IF_GT:				// 0x36:
		case IF_LE:				// 0x37:
			if (load) {
				set_register(&((info_register2T *) extraP)->reg1, (code >> 8) & 0xF);
				set_register(&((info_register2T *) extraP)->reg2,  code >> 12);
				assemblerP->m_jump.offset = (s2T) *codeP;
			}
			++codeP;
			extraP += sizeof(info_register2T);
			break;		
		case FILL_ARRAY_DATA:	// 0x26: 31t -> AA|op BBBBlo BBBBhi -> op vAA, +BBBBBBBB fill-array-data vAA, +BBBBBBBB (with supplemental data as specified below in "fill-array-data Format")
								// Note: The total number of code units for an instance of this table is (size * element_width + 1) / 2 + 4.
		{
			s4T offset;
			u2T	*dataP;
			u4T	data;

			offset = (s4T) getU4(&codeP);
			dataP  = codeP + offset - 3;
			assert(*dataP == 0x0300);			// 0x0300 identifying pseudo-opcode
			++dataP;
			width   = *dataP++;	// ushort element_widtdh - number of bytes in each element
			entries = getU4(&dataP);
			data    = entries * width;

			if (load) {
				set_register(&((info_fillT *) extraP)->reg1, code >> 8);
				((info_fillT *) extraP)->width   = width;	// ushort element_widtdh - number of bytes in each element
				((info_fillT *) extraP)->entries = entries; 
				memcpy( ((info_fillT *) extraP)->data, dataP, data);
			}
			extraP   += sizeof(info_fillT) + ((data + 7) & ~7);
			break;
		}
		case PACKED_SWITCH:		// 0x2b: 31t -> AA|op BBBBlo BBBBhi -> op vAA, +BBBBBBBB (with supplemental data as specified below in "packed-switch Format")
		{
			s4T offset;
			u2T	*dataP;
			u4T	value;
			u4T	branch;
			u4T	i;

			offset       = (s4T) getU4(&codeP);
			dataP        = codeP + offset - 3;
			assert(*dataP == 0x0100);
			++dataP;
			entries = *dataP++;
			value   = getU4(&dataP);
			if (load) {
				assemblerP->m_pcode = SPARSE_SWITCH;	// Normalize
			}
			for (i = 0; i < entries; ++i) {
				branch = getU4(&dataP);
				if (load) {
					if (i) {
						assemblerP[0] = assemblerP[-1];
					}
					assemblerP->info.P  = extraP;
					set_register(&((info_valueT *) extraP)->reg1, code >> 8);
					((info_valueT *) extraP)->value     = value;
					assemblerP->m_jump.offset = branch;
				}
				extraP += sizeof(info_valueT);
				++assemblerP;
				++value;
			}
			continue;
		}
		case SPARSE_SWITCH:		// 0x2c: 31t -> AA|op BBBBlo BBBBhi -> op vAA, +BBBBBBBB  (with supplemental data as specified below in "sparse-switch Format")
		{
			s4T offset;
			u2T	*dataP;
			u2T	*branchP;
			u4T	value, branch;
			u4T	i;

/*
			{
				u2T	*code1P;

				code1P = start_codeP;
				for (; code1P < end_codeP; ++code1P) {
					if (code1P[0] == 0x200) {
						code1P = code1P;
			}	}	}
*/
			offset       = (s4T) getU4(&codeP);
			dataP        = codeP + offset - 3;		// codeP has moved forward three
			assert(*dataP == 0x0200);
			++dataP;
			entries = *dataP++;
			branchP = dataP + (entries << 1);

			for (i = 0; i < entries; ++i) {
				value  = getU4(&dataP);
				branch = getU4(&branchP);
				if (load) {
					if (i) {
						assemblerP[0] = assemblerP[-1];
					}
					assemblerP->info.P                 = extraP;
					set_register(&((info_valueT *) extraP)->reg1, code >> 8);
					((info_valueT *) extraP)->value    = value;
					assemblerP->m_jump.offset          = branch;
				}
				extraP += sizeof(info_valueT);
				++assemblerP;
			}
			continue;
		}
		case FILLED_NEW_ARRAY:	// 0x25: 3rc -> AA|op BBBB CCCC -> op {vCCCC .. vNNNN}, type@BBBB (NNNN = AA + CCCC - 1}
		{
			info_registersT *infoP;
			u4T				a;

			a = code >> 8;
			infoP = (info_registersT *) extraP;

			if (load) {
				u4T c = codeP[1];
				u4T	i;

				set_type(&(infoP->last), codeP[0]);
				infoP->size = a;
				for (i = 0; i < a; ++i) {
					set_register(infoP->reg + i, c);
					++c;
			}	}
			codeP  += 2;
			extraP += ((sizeof(info_registersT) + ((a - 1) * sizeof(registerT)) + 7) & ~7);
			break;
		}
		case INVOKE_VIRTUAL_RANGE: // 0x74: 3rc -> AA|op BBBB CCCC -> op {vCCCC .. vNNNN}, meth@BBBB
		case INVOKE_SUPER_RANGE://0x75:
		case INVOKE_DIRECT_RANGE://0x76:
		case INVOKE_STATIC_RANGE://0x77:
		case INVOKE_INTERFACE_RANGE://0x78:
		{
			info_registersT *infoP;
			u4T				a;

			a = code >> 8;
			infoP = (info_registersT *) extraP;

			if (load) {
				u4T c = codeP[1];
				u4T	i;

				set_method(&(infoP->last), codeP[0]);
				infoP->size = a;
				for (i = 0; i < a; ++i) {
					set_register(infoP->reg + i, c);
					++c;
			}	}
			codeP  += 2;
			extraP += ((sizeof(info_registersT) + ((a - 1) * sizeof(registerT)) + 7) & ~7);
			break;
		}
		case FILL_ARRAY:		// 0x24: 35c -> B|A|op CCCC G|F|E|D -> [B=5] op {vD, vE, vF, vG, vA}, type@CCCC
						//				 [B=4] op {vD, vE, vF, vG}, type@CCCC
						//				 [B=3] op {vD, vE, vF}, type@CCCC
						//				 [B=2] op {vD, vE}, type@CCCC
						//				 [B=1] op {vD}, type@CCCC
						//				 [B=0] op {}, type@CCCC
		{	
			info_registersT	*infoP;
			u2T	b;
			u2T	c;
			u2T	code1;

			b       = code >> 12;
			infoP   = (info_registersT *) extraP;
			extraP += ((sizeof(info_registersT) + ((b - 1) * sizeof(registerT)) + 7) & ~7);
			assert(b < 6);
			c     = *codeP++;
			code1 = *codeP++;

			if (load) {
				infoP->size = b;									// B
				set_type(&(infoP->last), c);						// CCCC
				if (!b) {
					break;
				}
				set_register(infoP->reg, code1 & 0xF);			// DD
				if (!--b) {
					break;
				}
				code1 >>= 4;
				set_register(infoP->reg + 1, code1 & 0xF);		// EE
				if (!--b) {
					break;
				}
				code1 >>= 4;
				set_register(infoP->reg + 2, code1 & 0xF);		// FF
				if (!--b) {
					break;
				}
				code1 >>= 4;
				set_register(infoP->reg + 3, code1);				// GG
				if (!--b) {
					break;
				}
				set_register(infoP->reg + 4, (code >> 8) & 0xF);	// AA
			}
			break;
		}
		case INVOKE_VIRTUAL:	// 0x6e: [B=5] op {vD, vE, vF, vG, vA}, meth@CCCC
		case INVOKE_SUPER:		// 0x6f: [B=4] op {vD, vE, vF, vG}, meth@CCCC
		case INVOKE_DIRECT:		// 0x70: [B=3] op {vD, vE, vF}, meth@CCCC
		case INVOKE_STATIC:		// 0x71: [B=2] op {vD, vE}, meth@CCCC
		case INVOKE_INTERFACE:	// 0x72: [B=1] op {vD}, meth@CCCC
						//				 [B=0] op {}, meth@CCCC
		{	
			info_registersT	*infoP;
			u2T	b;
			u2T	c;
			u2T	code1;

			b       = code >> 12;
			infoP   = (info_registersT *) extraP;
			extraP += ((sizeof(info_registersT) + ((b - 1) * sizeof(registerT)) + 7) & ~7);
			assert(b < 6);
			c     = *codeP++;
			code1 = *codeP++;

			if (load) {
				infoP->size = b;									// B
				set_method(&(infoP->last), c);						// CCCC
				if (!b) {
					break;
				}
				set_register(infoP->reg, code1 & 0xF);			// DD
				if (!--b) {
					break;
				}
				code1 >>= 4;
				set_register(infoP->reg + 1, code1 & 0xF);		// EE
				if (!--b) {
					break;
				}
				code1 >>= 4;
				set_register(infoP->reg + 2, code1 & 0xF);		// FF
				if (!--b) {
					break;
				}
				code1 >>= 4;
				set_register(infoP->reg + 3, code1);				// GG
				if (!--b) {
					break;
				}
				set_register(infoP->reg + 4, (code >> 8) & 0xF);	// AA
			}
			break;
		}
		case BAD_3E:			// 0x3e: Bad operation
		case BAD_3F:			// 0x3f:
		case BAD_40:			// 0x40:
		case BAD_41:			// 0x41:
		case BAD_42:			// 0x42:
		case BAD_43:			// 0x43:
		case BAD_73:			// 0x73:
		case BAD_79:			// 0x79:
		case BAD_7A:			// 0x7a:
		case BAD_E3:			// 0xe3:
		case BAD_E4:			// 0xe4:
		case BAD_E5:			// 0xe5:
		case BAD_E6:			// 0xe6:
		case BAD_E7:			// 0xe7:
		case BAD_E8:			// 0xe8:
		case BAD_E9:			// 0xe9:
		case BAD_EA:			// 0xea:
		case BAD_EB:			// 0xeb:
		case BAD_EC:			// 0xec:
		case BAD_ED:			// 0xed:
		case BAD_EE:			// 0xee:
		case BAD_EF:			// 0xef:
		case BAD_F0:			// 0xf0:
		case BAD_F1:			// 0xf1:
		case BAD_F2:			// 0xf2:
		case BAD_F3:			// 0xf3:
		case BAD_F4:			// 0xf4:
		case BAD_F5:			// 0xf5:
		case BAD_F6:			// 0xf6:
		case BAD_F7:			// 0xf7:
		case BAD_F8:			// 0xf8:
		case BAD_F9:			// 0xf9:
		case BAD_FA:			// 0xfa:
		case BAD_FB:			// 0xfb:
		case BAD_FC:			// 0xfc:
		case BAD_FD:			// 0xfd:
		case BAD_FE:			// 0xfe:
		case BAD_FF:			// 0xff:
			Cdexerror::error("Illegal instruction code");
		}
		++assemblerP;
	}
	if (codeP != end_codeP) {
		Cdexerror::error("Code length mismatch");
	}

	*assemblerPP = assemblerP;
	*extraPP     = extraP;
}

/* static */ void
Ccode_item::scan(u4T offset)
{
	Ccode_item  *code_itemP;
	u2T			*start_codeP;
	u2T			*end_codeP;
	Ctry_item	*start_try_itemP;
	Ctry_item	*end_try_itemP;
	Ctry_item	*try_itemP;
	u4T			start_addr, end_addr, type_idx, handler_addr;
	u4T			size;
	s4T			i, size1;
	u4T			assemblers;
	int			load, extra;
	int			catchall;
	u1T			*P;
	codeT		*assemblerP;
	u1T			*extraP;

	check_offset(offset);
	code_itemP = (Ccode_item *) (g_imageP + offset);

	g_functionP->m_registers = code_itemP->registers_size;
	g_functionP->m_incoming  = code_itemP->ins_size;
	g_functionP->m_outgoing  = code_itemP->outs_size;
	g_functionP->m_tries     = code_itemP->tries_size;

	if (code_itemP->debug_info_off) {
		Cdebug_info_item::scan(code_itemP->debug_info_off);
	}
	size        = code_itemP->insns_size;
	start_codeP = code_itemP->insns;
	end_codeP   = start_codeP + size;

	if (size & 1) {
		start_try_itemP = (Ctry_item *) (end_codeP + 1);
	} else {
		start_try_itemP = (Ctry_item *) end_codeP;
	}
	Csort::clear_sort();
	end_try_itemP = start_try_itemP + code_itemP->tries_size;
	for (try_itemP = start_try_itemP; try_itemP < end_try_itemP; ++try_itemP) {
		start_addr = try_itemP->start_addr; 
		end_addr   = try_itemP->insn_count + start_addr;

		P = ((u1T *) end_try_itemP) + try_itemP->handler_off;
		size1 = decode_sleb128(&P);
		if (size1 > 0) {
			catchall = 0;
		} else {
			catchall = 1;
			if (size1 < 0) {
				size1 = -size1;
		}	}
		for (i = 0; i < size1; ++i) {
			type_idx     = decode_uleb128(&P);
			handler_addr = decode_uleb128(&P);
			Csort::append_exception(start_addr, end_addr, i, type_idx, handler_addr);
		}
		if (catchall) {
			handler_addr = decode_uleb128(&P);
			Csort::append_exception(start_addr, end_addr, i, NO_INDEX, handler_addr);
	}	}

	Csort::sort();
	assemblerP = 0;
	extraP     = 0;
	for (load = 0; ; ++load) {
		load_code(start_codeP, size, &assemblerP, &extraP, load);
		if (load) {
			break;
		}

		g_functionP->m_assemblers = assemblers = assemblerP - (codeT *) 0;
		extra                     = extraP     - (u1T *)   0;

		g_functionP->m_assemblersP = assemblerP = (codeT *) Xmalloc(sizeof(codeT) * assemblers + extra);
		extraP      = (u1T *) (assemblerP + assemblers);
	}
	assert(assemblers == (u4T) (assemblerP - g_functionP->m_assemblersP));
	assert((extraP - (u1T *) (g_functionP->m_assemblersP + assemblers)) == extra);   
	if (g_max_function_lth < assemblers) {
		g_max_function_lth = assemblers;
	}
	fixup_branches(g_functionP->m_assemblersP, assemblerP);

};

/* static */ void
Cencoded_method::fixup(u1T **PP)
{
	u4T method_idx_diff;
	u4T access_flags;
	u4T code_off;

	method_idx_diff     = decode_uleb128(PP);
	access_flags        = decode_uleb128(PP);
	code_off            = decode_uleb128(PP);
	if (code_off) {
		Ccode_item::fixup(code_off);
}	}

/* static */ void
Cclass_data_item::fixup(u4T offset)
{
	u1T	*P;
	u4T	static_fields_size;
	u4T	instance_fields_size;
	u4T	direct_methods_size;
	u4T virtual_methods_size;
	u4T	i;

	in_data_section(offset);
	P = g_imageP + offset;
	static_fields_size   = decode_uleb128(&P);
	instance_fields_size = decode_uleb128(&P);
	direct_methods_size  = decode_sleb128(&P);
	virtual_methods_size = decode_sleb128(&P);
	for (i = static_fields_size; i; --i) {
		Cencoded_field::fixup(&P);
	}
	for (i = instance_fields_size; i; --i) {
		Cencoded_field::fixup(&P);
	}
	for (i = direct_methods_size; i; --i) {
		Cencoded_method::fixup(&P);
	}
	for (i = virtual_methods_size; i; --i) {
		Cencoded_method::fixup(&P);
	}
};

/* static */ void
Cclass_data_item::scan(u4T offset)
{
	u1T	*P;
	u4T	static_fields;
	u4T	instance_fields;
	u4T	direct_functions;
	u4T virtual_functions;
	u4T	need;
	u4T	last_idx, idx, code_offset;
	encoded_fieldT	*static_fieldsP, *instance_fieldsP, *fieldP, *end_fieldsP;
	Cfunction		*end_methodsP;

	in_data_section(offset);
	P = g_imageP + offset;
	g_classP->m_static_fields     = static_fields     = decode_uleb128(&P);
	g_classP->m_instance_fields   = instance_fields   = decode_uleb128(&P);
	g_classP->m_direct_functions  = direct_functions  = decode_uleb128(&P);
	g_classP->m_virtual_functions = virtual_functions = decode_uleb128(&P);

	need = ((static_fields + instance_fields) * sizeof(encoded_fieldT)); 
	g_classP->m_static_fieldsP   = static_fieldsP   = (encoded_fieldT *) Xmalloc(need);
	g_classP->m_instance_fieldsP = instance_fieldsP = static_fieldsP + static_fields;

	last_idx = 0;
	fieldP            = static_fieldsP;
	for (end_fieldsP = fieldP + static_fields; fieldP < end_fieldsP; ++fieldP) {
		idx            = decode_uleb128(&P);
		idx           += last_idx;
		last_idx       = idx;
		if (g_sourceP->m_fields <= idx) {
			Cdexerror::error("Invalid class_data.static_field_idx");
		}
		fieldP->fieldP = g_sourceP->m_fieldsP + idx;
		fieldP->access_flags = decode_uleb128(&P);
	}

	last_idx = 0;
	fieldP            = instance_fieldsP;
	for (end_fieldsP = fieldP + instance_fields; fieldP < end_fieldsP; ++fieldP) {
		idx            = decode_uleb128(&P);
		idx           += last_idx;
		last_idx       = idx;
		if (g_sourceP->m_fields <= idx) {
			Cdexerror::error("Invalid class_data.instance_field_idx");
		}
		fieldP->fieldP = g_sourceP->m_fieldsP + idx;
		fieldP->access_flags = decode_uleb128(&P);
	}

	g_classP->m_functionsP = g_functionP  = new Cfunction[direct_functions + virtual_functions];

	last_idx = 0;
	for (end_methodsP = g_functionP + direct_functions; g_functionP < end_methodsP; ++g_functionP) {
		g_functionP->m_classP = g_classP;
		idx            = decode_uleb128(&P);
		idx           += last_idx;
		last_idx       = idx;
		if (g_sourceP->m_methods <= idx) {
			Cdexerror::error("Invalid class_data.direct_method_idx");
		}
		g_functionP->m_methodP      = g_sourceP->m_methodsP + idx;
		g_functionP->m_access_flags = decode_uleb128(&P);
		code_offset           = decode_uleb128(&P);
		if (code_offset) {
			Ccode_item::scan(code_offset);
		} else {
			g_functionP->m_assemblers  = 0;
			g_functionP->m_assemblersP = 0;
	}	}

	last_idx = 0;
	for (end_methodsP = g_functionP + virtual_functions; g_functionP < end_methodsP; ++g_functionP) {
		g_functionP->m_classP = g_classP;
		idx            = decode_uleb128(&P);
		idx           += last_idx;
		last_idx       = idx;
		if (g_sourceP->m_methods <= idx) {
			Cdexerror::error("Invalid class_data.virtual_method_idx");
		}
		g_functionP->m_methodP      = g_sourceP->m_methodsP + idx;
		g_functionP->m_access_flags = decode_uleb128(&P);
		code_offset           = decode_uleb128(&P);
		if (code_offset) {
			Ccode_item::scan(code_offset);
		} else {
			g_functionP->m_assemblers  = 0;
			g_functionP->m_assemblersP = 0;
	}	}
};

/* static */ void
Cclass_def_item::check_offset(u4T size, u4T offset)
{
	if (!size) {
		return;
	}
	if (offset < g_headerP->header_size) {
		Cdexerror::error("Invalid class_def_off -- in header");
	}
	if (offset & 3) {
		Cdexerror::error("Invalid class_defs_off -- not integer aligned");
}	}

/* static */ void
Cclass_def_item::fixup_all(u4T size, u4T offset)
{
	u4T				val;
	Cclass_def_item	*classdefP, *endP;

	check_offset(size, offset);
	classdefP = (Cclass_def_item *) (g_imageP + offset);
	for (endP = classdefP + size; classdefP < endP; ++classdefP) {
		fixup_uint((u1T *) &classdefP->class_idx);
		fixup_uint((u1T *) &classdefP->access_flags);
		fixup_uint((u1T *) &classdefP->superclass_idx);
		fixup_uint((u1T *) &classdefP->interfaces_off);
		Ctype_list::fixup(classdefP->interfaces_off);
		fixup_uint((u1T *) &classdefP->source_file_idx);
		fixup_uint((u1T *) &classdefP->annotations_off);
		val = classdefP->annotations_off;
		if (val) {
			Cannotations_directory_item::fixup(val);
		}
		fixup_uint((u1T *) &classdefP->class_data_off);
		val = classdefP->class_data_off;
		if (val) {
			Cclass_data_item::fixup(val);
		}
		fixup_uint((u1T *) &classdefP->static_values_off);
}	}

/* static */ void
Cclass_def_item::scan_all(u4T size, u4T offset)
{
	u4T				val;
	Cclass_def_item	*classdefP, *endP;
	stringT			**PP;
	Cclass			**tailPP;

	check_offset(size, offset);
	classdefP = (Cclass_def_item *) (g_imageP + offset);
	g_sourceP->m_classdefs  = size;

	tailPP = &(g_sourceP->m_classdefsP);
	for (endP = classdefP + size; classdefP < endP; ++classdefP) {
		val = classdefP->class_idx;
		*tailPP = g_classP = Cclass::locate(getType(val));
		tailPP  = &g_classP->m_nextP;

		g_classP->m_sourceP      = g_sourceP;
		g_classP->m_access_flags = classdefP->access_flags;

		val = classdefP->superclass_idx;
		g_classP->m_superclassP = getType(val);
		PP = 0;
		g_classP->m_interfaces = Ctype_list::scan(classdefP->interfaces_off, &PP, 0);
		if (g_classP->m_interfaces < 1) {
			g_classP->m_interfacesPP = 0;
		} else {
			g_classP->m_interfacesPP = PP = (stringT **) Xmalloc(((u1T *) PP) - ((u1T *) 0));
			Ctype_list::scan(classdefP->interfaces_off, &PP, 1);
			assert((g_classP->m_interfacesPP + g_classP->m_interfaces) == PP);
		}
		val = classdefP->source_file_idx;
		if (val == NO_INDEX) {
			g_classP->m_source_fileP = 0;
		} else {
			g_classP->m_source_fileP = Cfile::locate(getString(val));
		}

		val = classdefP->annotations_off;
		if (!val) {
			memset(&g_classP->m_annotations, 0, sizeof(g_classP->m_annotations));
		} else {
			Cannotations_directory_item::scan(&g_classP->m_annotations, val);
		}
		val = classdefP->class_data_off;
		if (val) {
			Cclass_data_item::scan(val);
		}
		val = classdefP->static_values_off;
		if (!val) {
			g_classP->m_initial_valuesP = 0;
		} else {
			u1T	*memoryP, *P;
			int	load, need;

			in_data_section(val);
			memoryP = 0;
			for (load = 0; ; ++load) {
				P = g_imageP + val;
				encoded_array(&P, &g_classP->m_initial_valuesP, &memoryP, load);
				if (load) {
					assert((memoryP - (u1T *) g_classP->m_initial_valuesP) == need);
					break;
				}
				need    = memoryP - (u1T *) 0;
				memoryP = (u1T *) Xmalloc(need);
			}
		}
		++classdefP;
	}
	*tailPP = 0;
	return;
}

/* static */ void
Cheader_item::fixup(void)
{
	g_headerP = (Cheader_item *) g_imageP;
	if (g_image_size < (int) sizeof(Cheader_item)) {
		Cdexerror::error("Too small for header");
	}
	if (strcmp(g_headerP->magic, DEX_FILE_MAGIC )) {
		Cdexerror::error("Invalid magic string");
	}
	g_data_end = g_headerP->data_off + g_headerP->data_size;

	switch (g_headerP->endian_tag) {
	case ENDIAN_CONSTANT:
		return;
	case REVERSE_ENDIAN_CONSTANT:
		break;
	default:
		Cdexerror::error("Invalid endian_tag");
	}
	fixup_uint((u1T *) &g_headerP->endian_tag);
	if (g_headerP->endian_tag != ENDIAN_CONSTANT) {
		Cdexerror::error("Endian reversal broken");
	}
	fixup_uint((u1T *) &g_headerP->checksum);
	fixup_uint((u1T *) &g_headerP->file_size);
	fixup_uint((u1T *) &g_headerP->header_size);
	fixup_uint((u1T *) &g_headerP->link_size);
	fixup_uint((u1T *) &g_headerP->link_off);
	fixup_uint((u1T *) &g_headerP->map_off);
	fixup_uint((u1T *) &g_headerP->string_ids_size);
	fixup_uint((u1T *) &g_headerP->string_ids_off);
	fixup_uint((u1T *) &g_headerP->type_ids_size);
	fixup_uint((u1T *) &g_headerP->type_ids_off);
	fixup_uint((u1T *) &g_headerP->proto_ids_size);
	fixup_uint((u1T *) &g_headerP->proto_ids_off);
	fixup_uint((u1T *) &g_headerP->field_ids_size);
	fixup_uint((u1T *) &g_headerP->field_ids_off);
	fixup_uint((u1T *) &g_headerP->method_ids_size);
	fixup_uint((u1T *) &g_headerP->method_ids_off);
	fixup_uint((u1T *) &g_headerP->class_defs_size);
	fixup_uint((u1T *) &g_headerP->class_defs_off);
	fixup_uint((u1T *) &g_headerP->data_size);
	fixup_uint((u1T *) &g_headerP->data_off);

	g_data_end = g_headerP->data_off + g_headerP->data_size;

	Cstring_id_item::fixup_all(g_headerP->string_ids_size, g_headerP->string_ids_off);
	Ctype_id_item::fixup_all(g_headerP->type_ids_size,     g_headerP->type_ids_off);
	Cproto_id_item::fixup_all(g_headerP->proto_ids_size,   g_headerP->proto_ids_off);
	Cfield_id_item::fixup_all(g_headerP->field_ids_size,   g_headerP->field_ids_off);
	Cmethod_id_item::fixup_all(g_headerP->method_ids_size, g_headerP->method_ids_off);
	Cclass_def_item::fixup_all(g_headerP->class_defs_size, g_headerP->class_defs_off);
}

/* static */ void
Cheader_item::scan(void)
{
	g_data_end = g_headerP->data_off + g_headerP->data_size;

	Cstring_id_item::scan_all(g_headerP->string_ids_size, g_headerP->string_ids_off);
	Ctype_id_item::scan_all(g_headerP->type_ids_size,     g_headerP->type_ids_off);
	Cproto_id_item::scan_all(g_headerP->proto_ids_size,   g_headerP->proto_ids_off);
	Cfield_id_item::scan_all(g_headerP->field_ids_size,   g_headerP->field_ids_off);
	Cmethod_id_item::scan_all(g_headerP->method_ids_size, g_headerP->method_ids_off);
	Cclass_def_item::scan_all(g_headerP->class_defs_size, g_headerP->class_defs_off);
}

int
Csource::load(const char *pathP, int dynamic)
{
	try {

		g_sourceP = this;
		g_sourceP->m_filenameP = pathP;
		Cheader_item::fixup();
		Cheader_item::scan();

		fprintf(stdoutF, "Opened %s\n<br>\n", pathP);

		if (g_verbose) {
			if (dynamic) {
				fputs("Dynamically processed", stderr);
			} else {
				fputs("Processed", stderr);
			}
			fprintf(stderr, " \"%s\"\n", pathP);
		}

		return(0);
	} catch (Cdexerror e) {
		const char *P;

		P = e.getError();
		fprintf(stdoutF, "Bad %s: %s\n<br>\n", pathP, P);
		if (g_verbose) {
			fprintf(stderr, "Errors in \"%s\"\n", pathP);
	}	}
	return(1);
}
